from typing import List

import datetime
import uuid

import jsonpatch


class SchemaVersion:

    def __init__(self, id_, date, changes):
        self.id = id_
        self.date = date
        self.changes = changes
        self.parents = []
        self.actions = []


def get_schema_changes(old: dict, new: dict) -> List[dict]:
    return [
        change
        for change in jsonpatch.make_patch(old, new)
        if not change['path'].startswith('/version/')
    ]


def get_new_schema_version(changes: List[dict]) -> dict:
    return SchemaVersion(
        str(uuid.uuid4()),
        datetime.datetime.now(datetime.timezone.utc),
        changes
    )



from typing import Any
from typing import Iterator
from typing import List

from spinta.components import Context
from spinta.components import Model
from spinta.datasets.components import Param
from spinta.dimensions.param.components import ParamBuilder
from spinta.dimensions.param.components import ResolvedParams


def _recursive_iter_params(
    context: Context,
    model: Model,
    params: List[Param],
    values: ResolvedParams,
) -> Iterator[ResolvedParams]:
    if len(params) == 1:
        yield values
    elif len(params) > 1:
        param = params[0]
        for source, formula in zip(param.sources, param.formulas):
            env = ParamBuilder(context)
            env.update(model=model, params=values, this=source)
            expr = env.resolve(formula)
            result: Iterator[Any] = env.execute(expr)
            if not isinstance(result, Iterator):
                result = iter([result])
            for value in result:
                yield from _recursive_iter_params(
                    context,
                    model,
                    params[1:],
                    {**values, param.name: value},
                )


def iterparams(context: Context, model: Model) -> Iterator[ResolvedParams]:
    values = {}
    params = model.external.params
    if params:
        raise NotImplementedError()
        yield from _recursive_iter_params(context, model, params, values)
    else:
        yield {}



from __future__ import annotations

from typing import List, Union, Any

import datetime
import dataclasses
from typing import Optional

import sqlalchemy as sa

from sqlalchemy.dialects.postgresql import UUID

from spinta import exceptions
from spinta.auth import authorized
from spinta.core.ufuncs import Env, ufunc
from spinta.core.ufuncs import Bind
from spinta.core.ufuncs import Expr
from spinta.exceptions import EmptyStringSearch
from spinta.exceptions import UnknownMethod
from spinta.exceptions import FieldNotInResource
from spinta.components import Action, Model, Property
from spinta.utils.data import take
from spinta.types.datatype import DataType
from spinta.types.datatype import Array
from spinta.types.datatype import File
from spinta.types.datatype import Object
from spinta.types.datatype import Ref
from spinta.types.datatype import String
from spinta.types.datatype import Integer
from spinta.types.datatype import Number
from spinta.types.datatype import DateTime
from spinta.types.datatype import Date
from spinta.types.datatype import PrimaryKey
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.components import BackendFeatures
from spinta.backends.postgresql.constants import TableType


class PgQueryBuilder(Env):
    backend: PostgreSQL

    def init(self, backend: PostgreSQL, table: sa.Table):
        return self(
            backend=backend,
            table=table,
            # TODO: Select list must be taken from params.select.
            select=None,
            joins={},
            from_=table,
            sort=[],
            limit=None,
            offset=None,
            aggregate=False,
        )

    def build(self, where):
        if self.select is None:
            self.call('select', Expr('select'))

        if self.aggregate:
            select = []
        else:
            select = [
                self.table.c['_id'],
                self.table.c['_revision'],
            ]
        for sel in self.select.values():
            items = sel.item if isinstance(sel.item, list) else [sel.item]
            for item in items:
                if item is not None and item not in select:
                    select.append(item)
        qry = sa.select(select)

        qry = qry.select_from(self.from_)

        if where is not None:
            qry = qry.where(where)

        if self.sort:
            qry = qry.order_by(*self.sort)

        if self.limit is not None:
            qry = qry.limit(self.limit)

        if self.offset is not None:
            qry = qry.offset(self.offset)

        return qry

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownMethod(expr=repr(expr(*args, **kwargs)), name=expr.name)

    def get_joined_table(self, prop: ForeignProperty) -> sa.Table:
        for fpr in prop.chain:
            if fpr.name in self.joins:
                continue

            if len(fpr.chain) > 1:
                ltable = self.joins[fpr.chain[-2].name]
            else:
                ltable = self.backend.get_table(fpr.left.model)
            lrkey = self.backend.get_column(ltable, fpr.left)

            rmodel = fpr.right.model
            rtable = self.backend.get_table(rmodel).alias()
            rpkey = self.backend.get_column(rtable, rmodel.properties['_id'])

            condition = lrkey == rpkey
            self.joins[fpr.name] = rtable
            self.from_ = self.from_.outerjoin(rtable, condition)

        return self.joins[fpr.name]


class ForeignProperty:

    def __init__(
        self,
        fpr: Optional[ForeignProperty],
        left: Property,
        right: Property,
    ):
        if fpr is None:
            self.name = left.place
            self.chain = [self]
        else:
            self.name = fpr.name + '->' + left.place
            self.chain = fpr.chain + [self]

        self.left = left
        self.right = right

    def __repr__(self):
        return (
            'ForeignProperty('
            f'{self.name}->{self.right.name}:{self.right.dtype.name}'
            ')'
        )

    @property
    def place(self):
        return '.'.join(
            [fpr.left.place for fpr in self.chain] +
            [self.right.place]
        )


class Func:
    pass


@dataclasses.dataclass
class Lower(Func):
    dtype: DataType = None


@dataclasses.dataclass
class Negative(Func):
    arg: Any


@dataclasses.dataclass
class Positive(Func):
    arg: Any


class Star:
    pass


@dataclasses.dataclass
class Recurse(Func):
    args: List[Union[DataType, Func]] = None


@dataclasses.dataclass
class ReservedProperty(Func):
    dtype: DataType
    param: str


@ufunc.resolver(PgQueryBuilder, str, name='op')
def op_(env, arg: str):
    if arg == '*':
        return Star()
    else:
        raise NotImplementedError


@ufunc.resolver(PgQueryBuilder, Bind, Bind, name='getattr')
def getattr_(env, field, attr):
    if field.name in env.model.properties:
        prop = env.model.properties[field.name]
    else:
        raise FieldNotInResource(env.model, property=field.anem)
    return env.call('getattr', prop.dtype, attr)


@ufunc.resolver(PgQueryBuilder, Object, Bind, name='getattr')
def getattr_(env, dtype, attr):
    if attr.name in dtype.properties:
        return dtype.properties[attr.name].dtype
    else:
        raise FieldNotInResource(dtype, property=attr.name)


@ufunc.resolver(PgQueryBuilder, Array, Bind, name='getattr')
def getattr_(env, dtype, attr):
    return env.call('getattr', dtype.items.dtype, attr)


@ufunc.resolver(PgQueryBuilder, File, Bind, name='getattr')
def getattr_(env, dtype, attr):
    return ReservedProperty(dtype, attr.name)


@ufunc.resolver(PgQueryBuilder, Ref, Bind, name='getattr')
def getattr_(env, dtype, attr):
    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype.prop, prop)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Bind, name='getattr')
def getattr_(env, fpr, attr):
    prop = fpr.right.dtype.model.properties[attr.name]
    return ForeignProperty(fpr, fpr.right, prop)


@ufunc.resolver(PgQueryBuilder, Expr)
def select(env, expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())

    env.select = {}

    if args:
        for key, arg in args:
            selected = env.call('select', arg)
            if selected is not None:
                env.select[key] = selected
    else:
        env.call('select', Star())

    assert env.select, args


@ufunc.resolver(PgQueryBuilder, Star)
def select(env, arg: Star) -> None:
    for prop in take(env.model.properties).values():
        # if authorized(env.context, prop, (Action.GETALL, Action.SEARCH)):
        # TODO: This line above should come from a getall(request),
        #       because getall can be used internally for example for
        #       writes.
        env.select[prop.place] = env.call('select', prop.dtype)


@ufunc.resolver(PgQueryBuilder, Bind)
def select(env, arg):
    if arg.name == '_type':
        return Selected(None, env.model.properties['_type'])
    prop = _get_property_for_select(env, arg.name)
    return env.call('select', prop.dtype)


def _get_property_for_select(env: PgQueryBuilder, name: str):
    prop = env.model.properties.get(name)
    if prop and authorized(env.context, prop, Action.SEARCH):
        return prop
    else:
        raise FieldNotInResource(env.model, property=name)


@dataclasses.dataclass
class Selected:
    item: Any
    prop: Property = None


@ufunc.resolver(PgQueryBuilder, DataType)
def select(env, dtype):
    table = env.backend.get_table(env.model)
    if dtype.prop.list is None:
        column = env.backend.get_column(table, dtype.prop, select=True)
    else:
        # XXX: Probably if dtype.prop is in a nested list, we need to get the
        #      first list. Because if I remember correctly, only first list is
        #      stored on the main table.
        column = env.backend.get_column(table, dtype.prop.list, select=True)
    return Selected(column, dtype.prop)


@ufunc.resolver(PgQueryBuilder, Object)
def select(env, dtype):
    columns = []
    for prop in take(dtype.properties).values():
        sel = env.call('select', prop.dtype)
        if isinstance(sel.item, list):
            columns += sel.item
        else:
            columns += [sel.item]
    return Selected(columns, dtype.prop)


@ufunc.resolver(PgQueryBuilder, File)
def select(env, dtype):
    table = env.backend.get_table(env.model)
    columns = [
        table.c[dtype.prop.place + '._id'],
        table.c[dtype.prop.place + '._content_type'],
        table.c[dtype.prop.place + '._size'],
    ]
    same_backend = dtype.backend.name == dtype.prop.model.backend.name
    if same_backend or BackendFeatures.FILE_BLOCKS in dtype.backend.features:
        columns += [
            table.c[dtype.prop.place + '._bsize'],
            table.c[dtype.prop.place + '._blocks'],
        ]
    return Selected(columns, dtype.prop)


@ufunc.resolver(PgQueryBuilder, ReservedProperty)
def select(env, prop):
    return env.call('select', prop.dtype, prop.param)


@ufunc.resolver(PgQueryBuilder, File, str)
def select(env, dtype, leaf):
    table = env.backend.get_table(env.model)
    if leaf == '_content':
        return env.call('select', dtype)
    else:
        column = table.c[dtype.prop.place + '.' + leaf]
        return Selected(column, dtype.prop)


@ufunc.resolver(PgQueryBuilder, Ref)
def select(env, dtype):
    table = env.backend.get_table(env.model)
    column = table.c[dtype.prop.place + '._id']
    return Selected(column, dtype.prop)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def select(env: PgQueryBuilder, fpr: ForeignProperty):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    column = column.label(fpr.place)
    return Selected(column, fpr.right)


@ufunc.resolver(PgQueryBuilder, int)
def limit(env, n):
    env.limit = n


@ufunc.resolver(PgQueryBuilder, int)
def offset(env, n):
    env.offset = n


@ufunc.resolver(PgQueryBuilder, Expr, name='and')
def and_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    if len(args) > 1:
        return sa.and_(*args)
    elif args:
        return args[0]


@ufunc.resolver(PgQueryBuilder, Expr, name='or')
def or_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return env.call('or', args)


@ufunc.resolver(PgQueryBuilder, list, name='or')
def or_(env, args):
    if len(args) > 1:
        return sa.or_(*args)
    elif args:
        return args[0]


@ufunc.resolver(PgQueryBuilder)
def count(env):
    env.aggregate = True
    env.select = {
        'count()': Selected(sa.func.count().label('count()')),
    }


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


@ufunc.resolver(PgQueryBuilder, Bind, object, names=COMPARE)
def compare(env, op, field, value):
    prop = _get_from_flatprops(env.model, field.name)
    return env.call(op, prop.dtype, value)


def _get_from_flatprops(model: Model, prop: str):
    if prop in model.flatprops:
        return model.flatprops[prop]
    else:
        raise exceptions.FieldNotInResource(model, property=prop)


@ufunc.resolver(PgQueryBuilder, DataType, object, names=COMPARE)
def compare(env, op, dtype, value):
    raise exceptions.InvalidValue(dtype, op=op, arg=type(value).__name__)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType, object, names=COMPARE)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DataType,
    value: Any,
):
    raise exceptions.InvalidValue(dtype, op=op, arg=type(value).__name__)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, object, names=COMPARE)
def compare(env, op: str, fpr: ForeignProperty, value: Any):
    return env.call(op, fpr, fpr.right.dtype, value)


@ufunc.resolver(PgQueryBuilder, DataType, type(None))
def eq(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType, type(None))
def eq(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: type(None),
):
    table = env.get_joined_table(fpr)
    column = env.backend.get_column(table, fpr.right)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, PrimaryKey, str)
def eq(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: type(None),
):
    table = env.backend.get_table(fpr.left.model)
    column = env.backend.get_column(table, fpr.left)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


def _ensure_non_empty(op, s):
    if s == '':
        raise EmptyStringSearch(op=op)


@ufunc.resolver(PgQueryBuilder, String, str, names=[
    'eq', 'startswith', 'contains',
])
def compare(env, op, dtype, value):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, String, str, names=[
    'eq', 'startswith', 'contains',
])
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: String,
    value: str,
):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, str, names=[
    'eq', 'startswith', 'contains',
])
def compare(env, op, dtype, value):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = env.backend.get_column(env.table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(PgQueryBuilder, (Integer, Number), (int, float), names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, (Integer, Number), (int, float), names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: Union[Integer, Number],
    value: Union[int, float],
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, DateTime, str, names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.datetime.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DateTime, str, names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DateTime,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    value = datetime.datetime.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Date, str, names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.date.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Date, str, names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: Date,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    value = datetime.date.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, String)
def lower(env, dtype):
    return Lower(dtype)


@ufunc.resolver(PgQueryBuilder, Recurse)
def lower(env, recurse):
    return Recurse([env.call('lower', arg) for arg in recurse.args])


@ufunc.resolver(PgQueryBuilder, Lower, str, names=[
    'eq', 'startswith', 'contains',
])
def compare(env, op, fn, value):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = env.backend.get_column(env.table, fn.dtype.prop)
    column = sa.func.lower(column)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, fn.dtype.prop, cond)


def _sa_compare(op, column, value):
    if op == 'eq':
        return column == value

    if op == 'lt':
        return column < value

    if op == 'le':
        return column <= value

    if op == 'gt':
        return column > value

    if op == 'ge':
        return column >= value

    if op == 'contains':
        if isinstance(column.type, UUID):
            column = column.cast(sa.String)
        return column.contains(value)

    if op == 'startswith':
        if isinstance(column.type, UUID):
            column = column.cast(sa.String)
        return column.startswith(value)
    raise NotImplementedError


def _prepare_condition(env: PgQueryBuilder, prop: Property, cond):
    if prop.list is None:
        return cond

    main_table = env.table
    list_table = env.backend.get_table(prop.list, TableType.LIST)
    subqry = (
        sa.select(
            [list_table.c._rid],
            distinct=list_table.c._rid,
        ).
        where(cond).
        alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry,
        main_table.c._id == subqry.c._rid,
    )
    return subqry.c._rid.isnot(None)


@ufunc.resolver(PgQueryBuilder, DataType, type(None))
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType, type(None))
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: type(None),
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, String, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, String, str)
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, (Integer, Number), (int, float))
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, (Integer, Number), (int, float))
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Union[Integer, Number],
    value: Union[int, float],
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, DateTime, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.datetime.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DateTime, str)
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DateTime,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    value = datetime.datetime.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, Date, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.date.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Date, str)
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Date,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    value = datetime.date.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, Array, (object, type(None)), names=[
    'eq', 'ne', 'lt', 'le', 'gt', 'ge', 'contains', 'startswith',
])
def compare(env, op, dtype, value):
    return env.call(op, dtype.items.dtype, value)


@ufunc.resolver(PgQueryBuilder, Lower, str)
def ne(env, fn, value):
    column = env.backend.get_column(env.table, fn.dtype.prop)
    column = sa.func.lower(column)
    return _ne_compare(env, fn.dtype.prop, column, value)


def _ne_compare(env: PgQueryBuilder, prop: Property, column, value):
    """Not equal operator is quite complicated thing and need explaining.

    If property is not defined within a list, just do `!=` comparison and be
    done with it.

    If property is in a list:

    - First check if there is at least one list item where field is not None
        (existance check).

    - Then check if there is no list items where field equals to given
        value.
    """

    if prop.list is None:
        return column != value

    main_table = env.backend.get_table(prop.model)
    list_table = env.backend.get_table(prop.list, TableType.LIST)

    # Check if at liest one value for field is defined
    subqry1 = (
        sa.select(
            [list_table.c._rid],
            distinct=list_table.c._rid,
        ).
        where(column != None).  # noqa
        alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry1,
        main_table.c._id == subqry1.c._rid,
    )

    # Check if given value exists
    subqry2 = (
        sa.select(
            [list_table.c._rid],
            distinct=list_table.c._rid,
        ).
        where(column == value).
        alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry2,
        main_table.c._id == subqry2.c._rid,
    )

    # If field exists and given value does not, then field is not equal to
    # value.
    return sa.and_(
        subqry1.c._rid != None,  # noqa
        subqry2.c._rid == None,
    )


FUNCS = [
    'lower',
    'upper',
]


@ufunc.resolver(PgQueryBuilder, Bind, names=FUNCS)
def func(env, name, field):
    prop = env.model.flatprops[field.name]
    return env.call(name, prop.dtype)


@ufunc.resolver(PgQueryBuilder, Bind)
def recurse(env, field):
    if field.name in env.model.leafprops:
        return Recurse([prop.dtype for prop in env.model.leafprops[field.name]])
    else:
        raise exceptions.FieldNotInResource(env.model, property=field.name)


@ufunc.resolver(PgQueryBuilder, Recurse, object, names=[
    'eq', 'ne', 'lt', 'le', 'gt', 'ge', 'contains', 'startswith',
])
def recurse(env, op, recurse, value):
    return env.call('or', [
        env.call(op, arg, value)
        for arg in recurse.args
    ])


@ufunc.resolver(PgQueryBuilder, Expr, name='any')
def any_(env, expr):
    args, kwargs = expr.resolve(env)
    op, field, *args = args
    if isinstance(op, Bind):
        op = op.name
    return env.call('or', [
        env.call(op, field, arg)
        for arg in args
    ])


@ufunc.resolver(PgQueryBuilder, Expr)
def sort(env, expr):
    args, kwargs = expr.resolve(env)
    env.sort = [
        env.call('sort', arg) for arg in args
    ]


@ufunc.resolver(PgQueryBuilder, Bind)
def sort(env, field):
    prop = _get_from_flatprops(env.model, field.name)
    return env.call('asc', prop.dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def sort(env: PgQueryBuilder, fpr: ForeignProperty):
    return env.call('asc', fpr, fpr.right.dtype)


@ufunc.resolver(PgQueryBuilder, Positive)
def sort(env, sign):
    return env.call('asc', sign.arg)


@ufunc.resolver(PgQueryBuilder, Negative)
def sort(env, sign):
    return env.call('desc', sign.arg)


@ufunc.resolver(PgQueryBuilder, DataType)
def asc(env, dtype):
    column = _get_sort_column(env, dtype.prop)
    return column.asc()


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def asc(env: PgQueryBuilder, fpr: ForeignProperty):
    return env.call('asc', fpr, fpr.right.dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType)
def asc(env: PgQueryBuilder, fpr: ForeignProperty, dtype: DataType):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return column.asc()


@ufunc.resolver(PgQueryBuilder, DataType)
def desc(env, dtype):
    column = _get_sort_column(env, dtype.prop)
    return column.desc()


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def desc(env: PgQueryBuilder, fpr: ForeignProperty):
    return env.call('desc', fpr, fpr.right.dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType)
def desc(env: PgQueryBuilder, fpr: ForeignProperty, dtype: DataType):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return column.desc()


@ufunc.resolver(PgQueryBuilder, Array, names=['asc', 'desc'])
def sort(env, name, dtype: Array):
    return env.call(name, dtype.items.dtype)


def _get_sort_column(env: PgQueryBuilder, prop: Property):
    column = env.backend.get_column(env.table, prop)

    if prop.list is None:
        return column

    main_table = env.table
    list_table = env.backend.get_table(prop.list, TableType.LIST)
    subqry = (
        sa.select(
            [list_table.c._rid, column.label('value')],
            distinct=list_table.c._rid,
        ).alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry,
        main_table.c._id == subqry.c._rid,
    )
    return subqry.c.value


@ufunc.resolver(PgQueryBuilder, Bind)
def negative(env, field) -> Negative:
    if field.name in env.model.properties:
        prop = env.model.properties[field.name]
    else:
        raise FieldNotInResource(env.model, property=field.name)
    return Negative(prop.dtype)


@ufunc.resolver(PgQueryBuilder, Bind)
def positive(env, field) -> Positive:
    if field.name in env.model.properties:
        prop = env.model.properties[field.name]
    else:
        raise FieldNotInResource(env.model, property=field.name)
    return Positive(prop.dtype)


@ufunc.resolver(PgQueryBuilder, DataType)
def negative(env, dtype) -> Negative:
    return Negative(dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def negative(env: PgQueryBuilder, fpr: ForeignProperty) -> Negative:
    return Negative(fpr)


@ufunc.resolver(PgQueryBuilder, DataType)
def positive(env, dtype) -> Positive:
    return Positive(dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def positive(env: PgQueryBuilder, fpr: ForeignProperty) -> Positive:
    return Positive(fpr)



from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.sqldump.components import SqlDump


@commands.wait.register(Context, SqlDump)
def wait(context: Context, backend: SqlDump, *, fail: bool = False) -> bool:
    # SqlDump is not a real backend.
    return True



from spinta import commands
from spinta.components import Context
from spinta.components import Model
from spinta.datasets.backends.frictionless.components import FrictionlessBackend


@commands.wipe.register(Context, Model, FrictionlessBackend)
def wipe(context: Context, model: Model, backend: FrictionlessBackend):
    raise NotImplementedError



import pathlib

from spinta.backends.components import Backend
from spinta.backends.components import BackendFeatures


class FileSystem(Backend):

    features = {
        BackendFeatures.WRITE,
    }

    path: pathlib.Path



import geoalchemy2 as ga
import sqlalchemy as sa

from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name
from spinta.components import Context
from spinta.types.geometry.components import Geometry


@commands.prepare.register(Context, PostgreSQL, Geometry)
def prepare(
    context: Context,
    backend: PostgreSQL,
    dtype: Geometry,
) -> sa.Column:
    prop = dtype.prop
    name = get_column_name(prop)
    kwargs = {
        'geometry_type': dtype.geometry_type,
        'srid': dtype.srid,
    }
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    column_type = ga.Geometry(**kwargs)
    return sa.Column(name, column_type, unique=dtype.unique)



import datetime
import hashlib

import msgpack


def get_ref_id(value):
    # XXX: deprecated use spinta.datasets.helpers.encode_primary_key
    if isinstance(value, list):
        value = list(value)
    else:
        value = [value]

    if any(v is None for v in value):
        return

    for i, v in enumerate(value):
        if isinstance(v, (datetime.datetime, datetime.date)):
            value[i] = v.isoformat()

    key = msgpack.dumps(value, use_bin_type=True)
    key = hashlib.sha1(key).hexdigest()

    return key



import datetime
import hashlib
import textwrap
import uuid

import itertools
import json
import logging
import pathlib

import pprintpp
import time
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import msgpack
import requests
import sqlalchemy as sa
import tqdm
from requests import HTTPError
from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import Exit
from typer import echo

from spinta import exceptions
from spinta import spyna
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.data import ModelRow
from spinta.cli.helpers.data import count_rows
from spinta.cli.helpers.data import ensure_data_dir
from spinta.cli.helpers.data import iter_model_rows
from spinta.cli.helpers.store import prepare_manifest
from spinta.client import RemoteClientCredentials
from spinta.client import get_access_token
from spinta.client import get_client_credentials
from spinta.components import Action
from spinta.components import Config
from spinta.components import Context
from spinta.components import Mode
from spinta.components import Model
from spinta.core.context import configure_context
from spinta.types.namespace import sort_models_by_refs
from spinta.utils.data import take
from spinta.utils.json import fix_data_for_json
from spinta.utils.nestedstruct import flatten
from spinta.utils.units import tobytes
from spinta.utils.units import toseconds

log = logging.getLogger(__name__)


def push(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Source manifest files to copy from"
    )),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Output data to a given location, by default outputs to stdout"
    )),
    credentials: str = Option(None, '--credentials', help=(
        "Credentials file, defaults to {config_path}/credentials.cfg"
    )),
    dataset: str = Option(None, '-d', '--dataset', help=(
        "Push only specified dataset"
    )),
    auth: str = Option(None, '-a', '--auth', help=(
        "Authorize as a client, defaults to {default_auth_client}"
    )),
    limit: int = Option(None, help=(
        "Limit number of rows read from each model"
    )),
    chunk_size: str = Option('1m', help=(
        "Push data in chunks (1b, 1k, 2m, ...), default: 1m"
    )),
    stop_time: str = Option(None, help=(
        "Stop pushing after given time (1s, 1m, 2h, ...), by default does not "
        "stops until all data is pushed"
    )),
    stop_row: int = Option(None, help=(
        "Stop after pushing n rows, by default does not stop until all data "
        "is pushed"
    )),
    state: pathlib.Path = Option(None, help=(
        "Save push state into a file, by default state is saved to "
        "{data_path}/push/{remote}.db SQLite database file"
    )),
    mode: Mode = Option('external', help=(
        "Mode of backend operation, default: external"
    )),
    dry_run: bool = Option(False, '--dry-run', help=(
        "Read data to be pushed, but do not push or write data to the "
        "destination."
    )),
    stop_on_error: bool = Option(False, '--stop-on-error', help=(
        "Exit immediately on first error."
    ))
):
    """Push data to external data store"""
    if chunk_size:
        chunk_size = tobytes(chunk_size)

    if stop_time:
        stop_time = toseconds(stop_time)

    context = configure_context(ctx.obj, manifests, mode=mode)
    store = prepare_manifest(context)
    config: Config = context.get('config')

    if credentials:
        credsfile = pathlib.Path(credentials)
        if not credsfile.exists():
            echo(f"Credentials file {credsfile} does not exit.")
            raise Exit(code=1)
    else:
        credsfile = config.credentials_file
    # TODO: Read client credentials only if a Spinta URL is given.
    creds = get_client_credentials(credsfile, output)

    if not state:
        ensure_data_dir(config.data_path / 'push')
        state = config.data_path / 'push' / f'{creds.remote}.db'

    manifest = store.manifest
    if dataset and dataset not in manifest.datasets:
        echo(str(exceptions.NodeNotFound(manifest, type='dataset', name=dataset)))
        raise Exit(code=1)

    ns = manifest.namespaces['']

    with context:
        client = auth or config.default_auth_client
        require_auth(context, client)
        context.attach('transaction', manifest.backend.transaction)

        backends = set()
        for backend in store.backends.values():
            backends.add(backend.name)
            context.attach(f'transaction.{backend.name}', backend.begin)
        for backend in manifest.backends.values():
            backends.add(backend.name)
            context.attach(f'transaction.{backend.name}', backend.begin)
        for dataset_ in manifest.datasets.values():
            for resource in dataset_.resources.values():
                if resource.backend and resource.backend.name not in backends:
                    backends.add(resource.backend.name)
                    context.attach(f'transaction.{resource.backend.name}', resource.backend.begin)
        for keymap in store.keymaps.values():
            context.attach(f'keymap.{keymap.name}', lambda: keymap)

        from spinta.types.namespace import traverse_ns_models

        models = traverse_ns_models(context, ns, Action.SEARCH, dataset)
        models = sort_models_by_refs(models)
        models = list(reversed(list(models)))
        counts = count_rows(
            context,
            models,
            limit,
            stop_on_error=stop_on_error,
        )

        if state:
            engine, metadata = _init_push_state(state, models)
            context.attach('push.state.conn', engine.begin)

        rows = iter_model_rows(
            context,
            models,
            counts,
            limit,
            stop_on_error=stop_on_error,
        )
        rows = _prepare_rows_for_push(rows)

        rows = tqdm.tqdm(rows, 'PUSH', ascii=True, total=sum(counts.values()))

        if stop_time:
            rows = _add_stop_time(rows, stop_time)

        if state:
            rows = _check_push_state(context, rows, metadata)

        if stop_row:
            rows = itertools.islice(rows, stop_row)

        rows = _push_to_remote_spinta(rows, creds, chunk_size, dry_run=dry_run)

        if state and not dry_run:
            rows = _save_push_state(context, rows, metadata)

        while True:
            try:
                next(rows)
            except StopIteration:
                break
            except:
                if stop_on_error:
                    raise
                log.exception("Error while reading data.")


class _PushRow:
    model: Model
    data: Dict[str, Any]
    rev: Optional[str]
    saved: bool = False

    def __init__(self, model: Model, data: Dict[str, Any]):
        self.model = model
        self.data = data
        self.rev = None
        self.saved = False


def _prepare_rows_for_push(rows: Iterable[ModelRow]) -> Iterator[_PushRow]:
    for model, row in rows:
        _id = row['_id']
        _type = row['_type']
        where = {
            'name': 'eq',
            'args': [
                {'name': 'bind', 'args': ['_id']},
                _id,
            ]
        }
        payload = {
            '_op': 'upsert',
            '_type': _type,
            '_id': _id,
            '_where': spyna.unparse(where),
            **{k: v for k, v in row.items() if not k.startswith('_')}
        }
        yield _PushRow(model, payload)


def _push_to_remote_spinta(
    rows: Iterable[_PushRow],
    creds: RemoteClientCredentials,
    chunk_size: int,
    *,
    dry_run: bool = False,
    stop_on_error: bool = False,
) -> Iterator[_PushRow]:
    echo(f"Get access token from {creds.server}")
    token = get_access_token(creds)

    session = requests.Session()
    session.headers['Content-Type'] = 'application/json'
    session.headers['Authorization'] = f'Bearer {token}'

    prefix = '{"_data":['
    suffix = ']}'
    slen = len(suffix)
    chunk = prefix
    ready = []

    for row in rows:
        data = fix_data_for_json(row.data)
        data = json.dumps(data, ensure_ascii=False)
        if ready and len(chunk) + len(data) + slen > chunk_size:
            yield from _send_and_receive(
                session,
                creds.server,
                ready,
                chunk + suffix,
                dry_run=dry_run,
                stop_on_error=stop_on_error,
            )
            chunk = prefix
            ready = []
        chunk += (',' if ready else '') + data
        ready.append(row)

    if ready:
        yield from _send_and_receive(
            session,
            creds.server,
            ready,
            chunk + suffix,
            dry_run=dry_run,
            stop_on_error=stop_on_error,
        )


def _get_row_for_error(rows: List[_PushRow]) -> str:
    size = len(rows)
    if size > 0:
        row = rows[0]
        data = pprintpp.pformat(row.data)
        return (
            f" Model {row.model.name},"
            f" items in chunk: {size},"
            f" first item in chunk:\n {data}"
        )
    else:
        return ''


def _send_and_receive(
    session: requests.Session,
    target: str,
    rows: List[_PushRow],
    data: str,
    *,
    dry_run: bool = False,
    stop_on_error: bool = False,
) -> Iterator[_PushRow]:
    if dry_run:
        recv = _send_data_dry_run(data)
    else:
        recv = _send_data(
            session,
            target,
            rows,
            data,
            stop_on_error=stop_on_error,
        )
    yield from _map_sent_and_recv(rows, recv)


def _send_data_dry_run(
    data: str,
) -> Optional[List[Dict[str, Any]]]:
    """Pretend data has been sent to a target location."""
    recv = json.loads(data)['_data']
    for row in recv:
        if '_id' not in row:
            row['_id'] = str(uuid.uuid4())
        row['_rev'] = str(uuid.uuid4())
    return recv


def _send_data(
    session: requests.Session,
    target: str,
    rows: List[_PushRow],
    data: str,
    *,
    stop_on_error: bool = False,
) -> Optional[List[Dict[str, Any]]]:
    data = data.encode('utf-8')

    try:
        resp = session.post(target, data=data)
    except IOError as e:
        if stop_on_error:
            raise
        log.error(
            (
                "Error when sending and receiving data.%s\n"
                "Error: %s"
            ),
            _get_row_for_error(rows),
            e,
        )
        return

    try:
        resp.raise_for_status()
    except HTTPError:
        if stop_on_error:
            raise
        log.error(
            (
                "Error when sending and receiving data.%s\n"
                "Server response (status=%s):\n%s"
            ),
            _get_row_for_error(rows),
            resp.status_code,
            textwrap.indent(pprintpp.pformat(resp.json()), '    '),
        )
        return

    return resp.json()['_data']


def _map_sent_and_recv(
    sent: List[_PushRow],
    recv: List[Dict[str, Any]],
) -> Iterator[_PushRow]:
    assert len(sent) == len(recv), (
        f"len(sent) = {len(sent)}, len(received) = {len(recv)}"
    )
    for sent_row, recv_row in zip(sent, recv):
        assert sent_row.data['_id'] == recv_row['_id'], (
            f"sent._id = {sent_row.data['_id']}, "
            f"received._id = {recv_row['_id']}"
        )
        yield sent_row


def _add_stop_time(rows, stop):
    start = time.time()
    for row in rows:
        yield row
        if time.time() - start > stop:
            break


def _init_push_state(
    file: pathlib.Path,
    models: List[Model],
) -> Tuple[sa.engine.Engine, sa.MetaData]:
    engine = sa.create_engine(f'sqlite:///{file}')
    metadata = sa.MetaData(engine)
    for model in models:
        table = sa.Table(
            model.name, metadata,
            sa.Column('id', sa.Unicode, primary_key=True),
            sa.Column('rev', sa.Unicode),
            sa.Column('pushed', sa.DateTime),
        )
        table.create(checkfirst=True)
    return engine, metadata


def _get_model_type(row: _PushRow) -> str:
    return row.data['_type']


def _check_push_state(
    context: Context,
    rows: Iterable[_PushRow],
    metadata: sa.MetaData,
):
    conn = context.get('push.state.conn')

    for model_type, group in itertools.groupby(rows, key=_get_model_type):
        table = metadata.tables[model_type]

        query = sa.select([table.c.id, table.c.rev])
        saved = {
            state[table.c.id]: state[table.c.rev]
            for state in conn.execute(query)
        }

        for row in group:
            _id = row.data['_id']

            rev = fix_data_for_json(take(row.data))
            rev = flatten([rev])
            rev = [[k, v] for x in rev for k, v in sorted(x.items())]
            rev = msgpack.dumps(rev, strict_types=True)
            rev = hashlib.sha1(rev).hexdigest()

            row.rev = rev
            row.saved = _id in saved

            if saved.get(_id) == row.rev:
                continue  # Nothing has changed.

            yield row


def _save_push_state(
    context: Context,
    rows: Iterable[_PushRow],
    metadata: sa.MetaData,
):
    conn = context.get('push.state.conn')
    for row in rows:
        table = metadata.tables[row.data['_type']]
        if row.saved:
            conn.execute(
                table.update().
                where(table.c.id == row.data['_id']).
                values(
                    id=row.data['_id'],
                    rev=row.rev,
                    pushed=datetime.datetime.now(),
                )
            )
        else:
            conn.execute(
                table.insert().
                values(
                    id=row.data['_id'],
                    rev=row.rev,
                    pushed=datetime.datetime.now(),
                )
            )
        yield row



from spinta import commands
from spinta.components import Context
from spinta.migrations import SchemaVersion
from spinta.backends.mongo.components import Mongo


@commands.freeze.register(Context, SchemaVersion, Mongo, object, object)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: Mongo,
    old: object,
    new: object,
):
    pass



import dataclasses

from spinta.components import Component
from spinta.core.enums import Access


@dataclasses.dataclass
class CommentGiven:
    access: str = None


@dataclasses.dataclass
class Comment(Component):
    id: str
    parent: str
    author: str
    access: Access
    # TODO: should be datetime
    created: str
    comment: str
    given: CommentGiven



from typing import List, Any

import datetime

from starlette.requests import Request

from spinta import commands
from spinta.auth import Token
from spinta.components import Context, Config
from spinta.components import UrlParams


class AccessLog:
    client: str = None
    method: str = None
    reason: str = None
    url: str = None
    buffer_size: int = 100
    format: str = None          # response format
    content_type: str = None    # request content-type header
    agent: str = None           # request user-agent header

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return None

    def create_message(
        self,
        *,
        ns: str = None,
        model: str = None,
        prop: str = None,
        txn: str = None,
        action: str = None,
        method: str = None,
        reason: str = None,
        id_: str = None,
        rev: str = None,
    ):
        message = {
            'agent': self.agent,
            'client': self.client,
            'rctype': self.content_type,
            'format': self.format,
            'action': action,
            'method': method or self.method,
            'url': self.url,
            'reason': reason or self.reason,
            'time': datetime.datetime.now(datetime.timezone.utc).isoformat(),
            'txn': txn,
            'ns': ns,
            'model': model,
            'prop': prop,
            'id': id_,
            'rev': rev,
        }
        return {k: v for k, v in message.items() if v is not None}

    def log(
        self,
        *,
        ns: str = None,
        model: str = None,
        prop: str = None,
        txn: str = None,
        action: str = None,
        method: str = None,
        reason: str = None,
        id_: str = None,
        rev: str = None,
    ):
        raise NotImplementedError


@commands.load.register(Context, AccessLog, Config)
def load(context: Context, accesslog: AccessLog, config: Config):
    accesslog.buffer_size = config.rc.get(
        'accesslog', 'buffer_size', required=True,
    )


@commands.load.register(Context, AccessLog, Token)
def load(context: Context, accesslog: AccessLog, token: Token):  # noqa
    accesslog.client = token.get_sub()


@commands.load.register(Context, AccessLog, Request)
def load(context: Context, accesslog: AccessLog, request: Request):  # noqa
    accesslog.method = request.method
    accesslog.url = str(request.url)
    accesslog.content_type = request.headers.get('content-type')
    accesslog.agent = request.headers.get('user-agent')


@commands.load.register(Context, AccessLog, UrlParams)
def load(context: Context, accesslog: AccessLog, params: UrlParams):  # noqa
    accesslog.format = params.format


def create_accesslog(
    context: Context,
    *,
    method: str = None,
    loaders: List[Any],
) -> AccessLog:
    config: Config = context.get('config')
    # XXX: Probably we should clone store.accesslog here, instead of creating
    #      completely new AccessLog instance.
    accesslog: AccessLog = config.AccessLog()
    accesslog.method = method
    for loader in loaders:
        commands.load(context, accesslog, loader)
    return accesslog



import csv
import re
from io import StringIO
from typing import Iterator
from typing import List
from typing import Tuple
from urllib.parse import parse_qsl
from urllib.parse import urlparse

import requests

gsheet_key_re = re.compile('/d/([^/]+)/')


def read_gsheets_manifest(path: str) -> Iterator[Tuple[str, List[str]]]:
    purl = urlparse(path)

    if match := gsheet_key_re.search(purl.path):
        key = match.group(1)
    else:
        raise RuntimeError(f"Unknown Google Sheet URL {path!r}.")

    url = f'https://docs.google.com/spreadsheets/d/{key}/gviz/tq'

    gid = None
    if purl.query:
        gid = dict(parse_qsl(purl.query)).get('gid')
    if gid is None and purl.fragment:
        gid = dict(parse_qsl(purl.fragment)).get('gid')

    params = {
        'tqx': 'out:csv',
        'gid': gid,
    }

    resp = requests.get(url, params)
    resp.raise_for_status()
    rows = csv.reader(StringIO(resp.text))
    for i, row in enumerate(rows, 2):
        yield str(i), row




from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.sql.components import Sql
from spinta.manifests.components import Manifest


@commands.prepare.register(Context, Sql, Manifest)
def prepare(context: Context, backend: Sql, manifest: Manifest):
    # XXX: Moved reflection to spinta/datasets/backends/sql/components:Sql.get_table
    # log.info(f"Reflecting database for {backend.name!r} backend, this might take time...")
    # backend.schema.reflect()
    pass



from typing import Optional, Union

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import File
from spinta.utils.schema import NotAvailable, NA


@commands.build_data_patch_for_write.register(Context, File)
def build_data_patch_for_write(
    context: Context,
    dtype: File,
    *,
    given: Optional[dict],
    saved: Optional[dict],
    insert_action: bool = False,
    update_action: bool = False,
) -> Union[dict, NotAvailable]:
    if insert_action or update_action:
        given = {
            '_id': given.get('_id', None) if given else None,
            '_content_type': given.get('_content_type', None) if given else None,
            '_content': given.get('_content', NA) if given else NA,
        }
    else:
        given = {
            '_id': given.get('_id', NA) if given else NA,
            '_content_type': given.get('_content_type', NA) if given else NA,
            '_content': given.get('_content', NA) if given else NA,
        }
    saved = {
        '_id': saved.get('_id', NA) if saved else NA,
        '_content_type': saved.get('_content_type', NA) if saved else NA,
        '_content': saved.get('_content', NA) if saved else NA,
    }
    given = {
        k: v for k, v in given.items()
        if v != saved[k]
    }
    given = {k: v for k, v in given.items() if v is not NA}
    return given or NA



import re

from spinta.types.datatype import DataType
from spinta.types.datatype import Date
from spinta.types.datatype import DateTime
from spinta.types.geometry.components import Geometry
from spinta.types.datatype import Integer
from spinta.types.datatype import Number

_time_unit_re = re.compile(r'^\d*[YMQWDHTSLUN]$')


def is_time_unit(unit: str) -> bool:
    return _time_unit_re.match(unit) is not None


# Borrowed from https://stackoverflow.com/a/3573731/475477
_prefix = '(Y|Z|E|P|T|G|M|k|h|da|d|c|m|µ|n|p|f|a|z|y)'
_unit = '(m|g|s|A|K|mol|cd|Hz|N|Pa|J|W|C|V|F|Ω|S|Wb|T|H|lm|lx|Bq|Gy|Sv|kat|l|L|B)'
_power = r'([⁺⁻]?[¹²³⁴⁵⁶⁷⁸⁹][⁰¹²³⁴⁵⁶⁷⁸⁹]*|\^[+-]?[1-9]\d*)'
_unit_and_prefix = '(' + _prefix + '?' + _unit + _power + '?|1)'
_multiplied = _unit_and_prefix + '(?:[⋅·*]' + _unit_and_prefix + ')*'
_with_denominator = _multiplied + '(?:/' + _multiplied + ')?'
_si_unit_re = re.compile(r'^\d*' + _with_denominator + '$')


def is_si_unit(unit: str) -> bool:
    return _si_unit_re.match(unit) is not None


def is_unit(dtype: DataType, unit: str) -> bool:
    if isinstance(dtype, (Date, DateTime)):
        return is_time_unit(unit)
    if isinstance(dtype, (Integer, Number, Geometry)):
        return is_si_unit(unit)
    return False



import base64
import datetime
import decimal
import pathlib
import uuid
from typing import Any
from typing import AsyncIterator
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional

from spinta import commands
from spinta import exceptions
from spinta.backends.components import Backend
from spinta.backends.components import SelectTree
from spinta.backends.helpers import get_model_reserved_props
from spinta.backends.helpers import select_keys
from spinta.backends.helpers import select_props
from spinta.backends.helpers import check_unknown_props
from spinta.backends.helpers import flat_select_to_nested
from spinta.backends.helpers import select_model_props
from spinta.backends.helpers import get_select_prop_names
from spinta.commands import gen_object_id
from spinta.commands import is_object_id
from spinta.commands import load_operator_value
from spinta.commands import prepare
from spinta.components import Action
from spinta.components import Context
from spinta.components import DataItem
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Node
from spinta.components import Property
from spinta.exceptions import ConflictingValue
from spinta.exceptions import NoItemRevision
from spinta.formats.components import Format
from spinta.types.datatype import Array
from spinta.types.datatype import Binary
from spinta.types.datatype import DataType
from spinta.types.datatype import Date
from spinta.types.datatype import Time
from spinta.types.datatype import DateTime
from spinta.types.datatype import File
from spinta.types.datatype import JSON
from spinta.types.datatype import Number
from spinta.types.datatype import Object
from spinta.types.datatype import PrimaryKey
from spinta.types.datatype import Ref
from spinta.types.datatype import String
from spinta.utils.data import take
from spinta.utils.schema import NA
from spinta.utils.schema import NotAvailable


@commands.prepare_for_write.register(Context, Model, Backend, dict)
def prepare_for_write(
    context: Context,
    model: Model,
    backend: Backend,
    patch: Dict[str, Any],
    *,
    action: Action,
) -> dict:
    # prepares model's data for storing in a backend
    backend = model.backend
    result = {}
    for name, value in patch.items():
        if not name.startswith('_'):
            prop = model.properties[name]
            value = commands.prepare_for_write(context, prop.dtype, backend, value)
        result[name] = value
    return result


@commands.prepare_for_write.register(Context, Property, Backend, object)
def prepare_for_write(
    context: Context,
    prop: Property,
    backend: Backend,
    value: Any,
    *,
    action: Action,
) -> Any:
    if prop.name in value:
        value[prop.name] = commands.prepare_for_write(
            context,
            prop.dtype,
            prop.dtype.backend,
            value[prop.name],
        )
    return value


@commands.prepare_for_write.register(Context, DataType, Backend, object)
def prepare_for_write(
    context: Context,
    dtype: DataType,
    backend: Backend,
    value: Any,
) -> Any:
    # prepares value for data store
    # for simple types - loaded native values should work
    # otherwise - override for this command if necessary
    return value


@commands.prepare_for_write.register(Context, Array, Backend, list)
def prepare_for_write(
    context: Context,
    dtype: Array,
    backend: Backend,
    value: List[Any],
) -> List[Any]:
    # prepare array and it's items for datastore
    return [
        commands.prepare_for_write(context, dtype.items.dtype, backend, v)
        for v in value
    ]


@commands.prepare_for_write.register(Context, Object, Backend, dict)
def prepare_for_write(
    context: Context,
    dtype: Object,
    backend: Backend,
    value: Dict[str, Any],
) -> Dict[str, Any]:
    # prepare objects and it's properties for datastore
    prepped = {}
    for k, v in value.items():
        prop = dtype.properties[k]
        prepped[k] = commands.prepare_for_write(context, prop.dtype, backend, v)
    return prepped


@prepare.register(Context, Backend, Property)
def prepare(context: Context, backend: Backend, prop: Property):
    return prepare(context, backend, prop.dtype)


@commands.simple_data_check.register(Context, DataItem, Model, Backend)
def simple_data_check(
    context: Context,
    data: DataItem,
    model: Model,
    backend: Backend,
) -> None:
    simple_model_properties_check(context, data, model, backend)


@commands.simple_data_check.register(Context, DataItem, Property, Backend)
def simple_data_check(
    context: Context,
    data: DataItem,
    prop: Property,
    backend: Backend,
) -> None:
    simple_data_check(
        context,
        data,
        prop.dtype,
        prop,
        backend,
        data.given[prop.name],
    )


def simple_model_properties_check(
    context: Context,
    data: DataItem,
    model: Model,
    backend: Backend,
) -> None:
    for name, prop in model.properties.items():
        simple_data_check(
            context,
            data,
            prop.dtype,
            prop,
            prop.dtype.backend,
            data.given.get(name, NA),
        )


@commands.simple_data_check.register(Context, DataItem, DataType, Property, Backend, object)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: DataType,
    prop: Property,
    backend: Backend,
    value: object,
) -> None:
    check_type_value(dtype, value)

    # Action.DELETE is ignore for qvarn compatibility reasons.
    # XXX: make `spinta` check for revision on Action.DELETE,
    #      implementers can override this command as they please.
    updating = data.action in (Action.UPDATE, Action.PATCH)
    if updating and '_revision' not in data.given:
        raise NoItemRevision(prop)


@commands.simple_data_check.register(Context, DataItem, Object, Property, Backend, dict)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: Object,
    prop: Property,
    backend: Backend,
    value: Dict[str, Any],
) -> None:
    for prop in dtype.properties.values():
        v = value.get(prop.name, NA)
        simple_data_check(context, data, prop.dtype, prop, prop.dtype.backend, v)


@commands.simple_data_check.register(Context, DataItem, Array, Property, Backend, list)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: Array,
    prop: Property,
    backend: Backend,
    value: list,
) -> None:
    dtype = dtype.items.dtype
    for v in value:
        simple_data_check(context, data, dtype, prop, dtype.backend, v)


@commands.complex_data_check.register(Context, DataItem, Model, Backend)
def complex_data_check(
    context: Context,
    data: DataItem,
    model: Model,
    backend: Backend,
) -> None:
    # XXX: Maybe Action.DELETE should also provide revision.
    updating = data.action in (Action.UPDATE, Action.PATCH)
    if updating and '_revision' not in data.given:
        raise NoItemRevision(model)

    if data.action in (Action.UPDATE, Action.PATCH, Action.DELETE):
        for k in ('_type', '_revision'):
            if k in data.given and data.saved[k] != data.given[k]:
                raise ConflictingValue(
                    model.properties[k],
                    given=data.given[k],
                    expected=data.saved[k],
                )

    complex_model_properties_check(context, model, backend, data)


def complex_model_properties_check(
    context: Context,
    model: Model,
    backend: Backend,
    data: DataItem,
) -> None:
    for name, prop in model.properties.items():
        # For datasets, property type is optional.
        # XXX: But I think, it should be mandatory.
        if prop.dtype is not None:
            given = data.given.get(name, NA)

            if prop.dtype.unique and given is not NA:
                commands.check_unique_constraint(
                    context,
                    data,
                    prop.dtype,
                    prop,
                    data.backend,
                    given,
                )
            complex_data_check(
                context,
                data,
                prop.dtype,
                prop,
                data.backend,
                given,
            )


@complex_data_check.register(Context, DataItem, DataType, Property, Backend, object)
def complex_data_check(
    context: Context,
    data: DataItem,
    dtype: DataType,
    prop: Property,
    backend: Backend,
    value: object,
):
    if data.action in (Action.UPDATE, Action.PATCH, Action.DELETE):
        for k in ('_type', '_revision'):
            if k in data.given and data.saved[k] != data.given[k]:
                raise ConflictingValue(
                    dtype,
                    given=data.given[k],
                    expected=data.saved[k],
                )


def check_type_value(dtype: DataType, value: object):
    if dtype.required and (value is None or value is NA):
        # FIXME: Raise a UserError
        raise Exception(f"{dtype.prop.name!r} is required for {dtype.prop.model.name!r}.")


@gen_object_id.register(Context, Backend, Node)
def gen_object_id(context: Context, backend: Backend, model: Node):
    return str(uuid.uuid4())


@is_object_id.register(Context, str)
def is_object_id(context: Context, value: str):
    # Collect all available backend/model combinations.
    # TODO: this probably should be cached at load time to increase performance
    #       of object id lookup.
    # XXX: other option would be to implement two pass URL parsing, to get
    #      information about model, and then call is_object_id directly with
    #      backend and model. That would be nice.
    store = context.get('store')
    candidates = set()
    manifest = store.manifest
    for model in manifest.objects.get('model', {}).values():
        candidates.add((model.backend.__class__, model.__class__))

    # Currently all models use UUID, but dataset models use id generated from
    # other properties that form an unique id.
    for Backend_, Model_ in candidates:
        backend = Backend_()
        model = Model_()
        model.name = ''
        if is_object_id(context, backend, model, value):
            return True


@is_object_id.register(Context, Backend, Model, str)
def is_object_id(context: Context, backend: Backend, model: Model, value: str):
    try:
        return uuid.UUID(value).version == 4
    except ValueError:
        return False


@is_object_id.register(Context, Backend, Model, object)
def is_object_id(context: Context, backend: Backend, model: Model, value: object):
    # returns false if object id is non string.
    # strings are handled by other command
    return False


@is_object_id.register(Context, type(None), Namespace, object)
def is_object_id(context: Context, backend: type(None), model: Namespace, value: object):
    # Namespaces do not have object id.
    return False


@commands.prepare_data_for_response.register(Context, Namespace, Format, dict)
def prepare_data_for_response(
    context: Context,
    ns: Namespace,
    fmt: Format,
    value: dict,
    *,
    action: Action,
    select: SelectTree,
    prop_names: List[str],
) -> dict:
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=value,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_model_props(
            ns.manifest.models['_ns'],
            prop_names,
            value,
            select,
            reserved=['_id'],
        )
    }


@commands.prepare_data_for_response.register(Context, Model, Format, dict)
def prepare_data_for_response(
    context: Context,
    model: Model,
    fmt: Format,
    value: dict,
    *,
    action: Action,
    select: SelectTree,
    prop_names: List[str],
) -> dict:
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=value,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_model_props(
            model,
            prop_names,
            value,
            select,
            get_model_reserved_props(action),
        )
    }


@commands.prepare_data_for_response.register(Context, Object, Format, dict)
def prepare_data_for_response(
    context: Context,
    dtype: Object,
    fmt: Format,
    value: dict,
    *,
    action: Action,
    select: List[str] = None,
) -> dict:
    select = flat_select_to_nested(select)
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            value,
            data=value,
            action=action,
            select=select,
        )
        for prop, value, select in _select_prop_props(
            dtype,
            value,
            select,
            ['_type', '_revision'],
        )
    }


@commands.prepare_data_for_response.register(Context, File, Format, dict)
def prepare_data_for_response(
    context: Context,
    dtype: File,
    fmt: Format,
    value: dict,
    *,
    action: Action,
    select: List[str] = None,
    property_: Property = None,
) -> dict:
    prop = dtype.prop
    reserved = ['_type', '_revision']

    if prop.name in value and value[prop.name]:
        check_unknown_props(prop, select, set(reserved) | set(value[prop.name]))
    else:
        check_unknown_props(prop, select, set(reserved))

    if prop.name in value and value[prop.name]:
        data = commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            value[prop.name],
            data=value,
            action=action,
            select=select,
        )
    else:
        data = {}

    for key, val, sel in select_keys(reserved, value, select):
        data[key] = val

    return data


def _select_prop_props(
    dtype: Object,
    value: dict,
    select: SelectTree,
    reserved: List[str],
):
    prop = dtype.prop
    if prop.name in value:
        check_unknown_props(prop, select, set(reserved) | set(value[prop.name]))
    else:
        check_unknown_props(prop, select, set(reserved))
    yield from select_props(
        prop,
        reserved,
        prop.model.properties,
        value,
        select,
    )
    if prop.name in value:
        yield from select_props(
            prop,
            value[prop.name].keys(),
            dtype.properties,
            value[prop.name],
            select,
        )


@commands.prepare_dtype_for_response.register(Context, Format, DataType, object)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: DataType,
    value: Any,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    assert isinstance(value, (str, int, float, bool, type(None))), (
        f"prepare_dtype_for_response must return only primitive, json "
        f"serializable types, {type(value)} is not a primitive data type, "
        f"model={dtype.prop.model!r}, dtype={dtype!r}"
    )
    return value


@commands.prepare_dtype_for_response.register(Context, Format, DataType, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: DataType,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return dtype.default


@commands.prepare_dtype_for_response.register(Context, Format, File, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: File,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return {
        '_id': None,
        '_content_type': None,
    }


@commands.prepare_dtype_for_response.register(Context, Format, File, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: File,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    data = {
        key: val
        for key, val, sel in select_keys(
            ['_id', '_content_type'],
            value,
            select,
        )
    }

    if isinstance(data.get('_id'), pathlib.Path):
        # _id on FileSystem backend is a Path.
        data['_id'] = str(data['_id'])

    # File content is returned only if explicitly requested.
    if select and '_content' in select:
        if '_content' in value:
            content = value['_content']
        else:
            prop = dtype.prop
            content = commands.getfile(context, prop, dtype, dtype.backend, data=value)

        if content is None:
            data['_content'] = None
        else:
            data['_content'] = base64.b64encode(content).decode()

    return data


@commands.prepare_dtype_for_response.register(Context, Format, DateTime, datetime.datetime)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: DateTime,
    value: datetime.datetime,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return value.isoformat()


@commands.prepare_dtype_for_response.register(Context, Format, Date, datetime.date)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Date,
    value: datetime.date,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if isinstance(value, datetime.datetime):
        return value.date().isoformat()
    else:
        return value.isoformat()


@commands.prepare_dtype_for_response.register(Context, Format, Time, datetime.time)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Time,
    value: datetime.time,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return value.isoformat()


@commands.prepare_dtype_for_response.register(Context, Format, Binary, bytes)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Binary,
    value: bytes,
    data: Dict[str, Any],
    *,
    action: Action,
    select: dict = None,
):
    return base64.b64encode(value).decode('ascii')


@commands.prepare_dtype_for_response.register(Context, Format, Ref, (dict, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Ref,
    value: Optional[Dict[str, Any]],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if value is None:
        value = {'_id': None}

    if select and select != {'*': {}}:
        names = get_select_prop_names(
            context,
            dtype,
            dtype.model.properties,
            action,
            select,
            reserved=['_id'],
        )
    else:
        names = ['_id']

    data = {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=data,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_props(
            dtype.model,
            names,
            dtype.model.properties,
            value,
            select,
        )
    }

    return data


@commands.prepare_dtype_for_response.register(Context, Format, Ref, str)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Ref,
    value: str,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    # FIXME: Backend should never return references as strings! References
    #        should always be dicts.
    return {'_id': value}


@commands.prepare_dtype_for_response.register(Context, Format, Object, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Object,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    names = get_select_prop_names(
        context,
        dtype,
        dtype.properties,
        action,
        select,
        # XXX: Not sure if auth=False should be here. But I'm too tiered to
        #      check why it does not work.
        auth=False,
    )
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=data,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_props(
            dtype.prop,
            names,
            dtype.properties,
            value,
            select,
        )
    }


@commands.prepare_dtype_for_response.register(Context, Format, Array, list)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Array,
    value: list,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
) -> list:
    return _prepare_array_for_response(
        context,
        dtype,
        fmt,
        value,
        data,
        action,
        select,
    )


@commands.prepare_dtype_for_response.register(Context, Format, Array, tuple)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Array,
    value: tuple,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
) -> list:
    return _prepare_array_for_response(
        context,
        dtype,
        fmt,
        value,
        data,
        action,
        select,
    )


def _prepare_array_for_response(
    context: Context,
    dtype: Array,
    fmt: Format,
    value: Iterable,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
) -> list:
    if dtype.items:
        return [
            commands.prepare_dtype_for_response(
                context,
                fmt,
                dtype.items.dtype,
                v,
                data=data,
                action=action,
                select=select,
            )
            for v in value
        ]
    else:
        return [v for v in value]


@commands.prepare_dtype_for_response.register(Context, Format, Array, type(None))
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Array,
    value: type(None),
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return []


@commands.prepare_dtype_for_response.register(Context, Format, JSON, object)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: JSON,
    value: object,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return value


@commands.prepare_dtype_for_response.register(Context, Format, JSON, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: JSON,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return None


@commands.prepare_dtype_for_response.register(Context, Format, Number, decimal.Decimal)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Number,
    value: decimal.Decimal,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return float(value)


@commands.unload_backend.register(Context, Backend)
def unload_backend(context: Context, backend: Backend):
    pass


@load_operator_value.register(Context, Backend, DataType, object)
def load_operator_value(
    context: Context,
    backend: Backend,
    dtype: DataType,
    value: object,
    *,
    query_params: dict,
):
    operator = query_params['name']
    # XXX: That is probably not a very reliable way of getting original operator
    #      name. Maybe at least this should be documented some how.
    operator_name = query_params.get('origin', operator)
    if operator in ('startswith', 'contains') and not isinstance(dtype, (String, PrimaryKey)):
        raise exceptions.InvalidOperandValue(dtype, operator=operator_name)
    if operator in ('gt', 'ge', 'lt', 'le') and isinstance(dtype, String):
        raise exceptions.InvalidOperandValue(dtype, operator=operator_name)


@commands.make_json_serializable.register(Model, dict)
def make_json_serializable(model: Model, value: dict) -> dict:
    return commands.make_json_serializable[Object, dict](model, value)


@commands.make_json_serializable.register(DataType, object)
def make_json_serializable(dtype: DataType, value: object):
    return value


@commands.make_json_serializable.register(DateTime, datetime.datetime)
def make_json_serializable(dtype: DateTime, value: datetime.datetime):
    return value.isoformat()


@commands.make_json_serializable.register(Date, datetime.date)
def make_json_serializable(dtype: Date, value: datetime.date):
    return value.isoformat()


@commands.make_json_serializable.register(Date, datetime.datetime)
def make_json_serializable(dtype: Date, value: datetime.datetime):
    return value.date().isoformat()


def _get_json_serializable_props(dtype: Object, value: dict):
    for k, v in value.items():
        if k in dtype.properties:
            yield dtype.properties[k], v
        else:
            Exception(f"Unknown {k} key in {dtype!r} object.")


@commands.make_json_serializable.register(Object, dict)
def make_json_serializable(dtype: Object, value: dict):
    return {
        prop.name: commands.make_json_serializable(prop.dtype, v)
        for prop, v in _get_json_serializable_props(dtype, value)
    }


@commands.make_json_serializable.register(Array, list)
def make_json_serializable(dtype: Array, value: list):
    return [make_json_serializable(dtype.items.dtype, v) for v in value]


@commands.update.register(Context, Property, DataType, Backend)
def update(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    return commands.update(
        context, prop.model, prop.model.backend,
        dstream=dstream,
        stop_on_error=stop_on_error,
    )


@commands.patch.register(Context, Property, DataType, Backend)
def patch(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    return commands.update(
        context, prop.model, prop.model.backend,
        dstream=dstream,
        stop_on_error=stop_on_error,
    )


@commands.delete.register(Context, Property, DataType, Backend)
def delete(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    dstream = _prepare_property_for_delete(prop, dstream)
    return commands.update(
        context, prop.model, prop.model.backend,
        dstream=dstream,
        stop_on_error=stop_on_error,
    )


async def _prepare_property_for_delete(
    prop: Property,
    dstream: AsyncIterator[DataItem],
) -> AsyncIterator[DataItem]:
    async for data in dstream:
        if data.patch:
            data.patch = {
                **{k: v for k, v in data.patch.items() if k.startswith('_')},
                prop.name: None,
            }
        yield data


@commands.cast_backend_to_python.register(Context, Model, Backend, dict)
def cast_backend_to_python(
    context: Context,
    model: Model,
    backend: Backend,
    data: dict,
) -> dict:
    return {
        k: commands.cast_backend_to_python(
            context,
            model.properties[k].dtype,
            backend,
            v,
        ) if k in model.properties else v
        for k, v in data.items()
    }


@commands.cast_backend_to_python.register(Context, Property, Backend, object)
def cast_backend_to_python(
    context: Context,
    prop: Property,
    backend: Backend,
    data: object,
) -> dict:
    return commands.cast_backend_to_python(context, prop.dtype, backend, data)


@commands.cast_backend_to_python.register(Context, DataType, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: DataType,
    backend: Backend,
    data: Any,
) -> Any:
    return data


@commands.cast_backend_to_python.register(Context, Object, Backend, dict)
def cast_backend_to_python(
    context: Context,
    dtype: Object,
    backend: Backend,
    data: Dict[str, Any],
) -> Dict[str, Any]:
    return {
        k: commands.cast_backend_to_python(
            context,
            dtype.properties[k].dtype,
            backend,
            v,
        ) if k in dtype.properties else v
        for k, v in data.items()
    }


@commands.cast_backend_to_python.register(Context, Array, Backend, list)
def cast_backend_to_python(
    context: Context,
    dtype: Array,
    backend: Backend,
    data: List[Any],
) -> List[Any]:
    return [
        commands.cast_backend_to_python(
            context,
            dtype.items.dtype,
            backend,
            v,
        )
        for v in data
    ]


@commands.cast_backend_to_python.register(Context, Binary, Backend, str)
def cast_backend_to_python(
    context: Context,
    dtype: Binary,
    backend: Backend,
    data: str,
) -> bytes:
    return base64.b64decode(data)



from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

from spinta import commands
from spinta import exceptions
from spinta import spyna
from spinta.auth import authorized
from spinta.backends import Backend
from spinta.backends.components import SelectTree
from spinta.backends.components import BackendOrigin
from spinta.components import Action
from spinta.components import Component
from spinta.components import Context
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Property
from spinta.types.datatype import DataType
from spinta.utils.data import take


def load_backend(
    context: Context,
    component: Component,
    name: str,
    origin: BackendOrigin,
    data: Dict[str, str]
) -> Backend:
    config = context.get('config')
    type_ = data.get('type')
    if not type_:
        raise exceptions.RequiredConfigParam(
            component,
            name=f'backends.{name}.type',
        )
    if type_ not in config.components['backends']:
        raise exceptions.BackendNotFound(component, name=type_)
    Backend_ = config.components['backends'][type_]
    backend: Backend = Backend_()
    backend.type = type_
    backend.name = name
    backend.origin = origin
    backend.config = data
    commands.load(context, backend, data)
    return backend


def get_select_tree(
    context: Context,
    action: Action,
    select: Optional[List[str]],
) -> SelectTree:
    select = _apply_always_show_id(context, action, select)
    if select is None and action in (Action.GETALL, Action.SEARCH):
        # If select is not given, select everything.
        select = {'*': {}}
    return flat_select_to_nested(select)


def _apply_always_show_id(
    context: Context,
    action: Action,
    select: Optional[List[str]],
) -> Optional[List[str]]:
    if action in (Action.GETALL, Action.SEARCH):
        config = context.get('config')
        if config.always_show_id:
            if select is None:
                return ['_id']
            elif '_id' not in select:
                return ['_id'] + select
    return select


def get_select_prop_names(
    context: Context,
    node: Union[Model, Property, DataType],
    props: Dict[str, Property],
    action: Action,
    select: SelectTree,
    *,
    # If False, do not check if client has access to this property.
    auth: bool = True,
    # Allowed reserved property names.
    reserved: List[str] = None,
) -> List[str]:
    known = set(reserved or []) | set(take(props))
    check_unknown_props(node, select, known)

    if select is None or '*' in select:
        return [
            p.name
            for p in props.values() if (
                not p.name.startswith('_') and
                not p.hidden and
                (not auth or authorized(context, p, action))
            )
        ]
    else:
        return list(select)


def select_model_props(
    model: Model,
    prop_names: List[str],
    value: dict,
    select: SelectTree,
    reserved: List[str],
) -> Iterator[Tuple[
    Union[Property, str],
    Any,
    SelectTree,
]]:
    yield from select_props(
        model,
        reserved,
        model.properties,
        value,
        select,
    )
    yield from select_props(
        model,
        prop_names,
        model.properties,
        value,
        select,
        reserved=False,
    )


T = TypeVar('T')


def select_props(
    node: Union[Namespace, Model, Property],
    keys: Iterable[str],
    props: Dict[str, Property],
    value: Dict[str, T],
    select: SelectTree,
    *,
    reserved: bool = True,
) -> Iterator[Tuple[
    Union[Property, str],
    T,
    SelectTree,
]]:
    for key, val, sel in select_keys(keys, value, select, reserved=reserved):
        prop = _select_prop(key, props, node)
        if prop:
            yield prop, val, sel


def select_only_props(
    node: Union[Namespace, Model, Property],
    keys: Iterable[str],
    props: Dict[str, Property],
    select: SelectTree,
    *,
    reserved: bool = True,
) -> Iterator[Tuple[
    Union[Property, str],
    SelectTree,
]]:
    for key, sel in select_only_keys(keys, select, reserved=reserved):
        prop = _select_prop(key, props, node)
        if prop:
            yield prop, sel


def _select_prop(
    key: str,
    props: Dict[str, Property],
    node: Union[Namespace, Model, Property],
) -> Optional[Property]:
    if key not in props:
        # FIXME: We should check select list at the very beginning of
        #        request, not when returning results.
        raise exceptions.FieldNotInResource(node, property=key)
    prop = props[key]
    if not prop.hidden:
        return prop


def select_keys(
    keys: Iterable[str],
    value: Dict[str, T],
    select: SelectTree,
    *,
    reserved: bool = True,
) -> Iterator[Tuple[
    str,
    T,
    SelectTree,
]]:
    for key, sel in select_only_keys(keys, select, reserved=reserved):
        if select is None and key not in value:
            # Omit all keys if they are not present in value, this is a common
            # case in PATCH requests.
            continue

        if key in value:
            val = value[key]
        else:
            val = None

        yield key, val, sel


def select_only_keys(
    keys: Iterable[str],
    select: SelectTree,
    *,
    reserved: bool = True,
) -> Iterator[Tuple[
    str,
    SelectTree,
]]:
    for key in keys:
        if reserved is False and key.startswith('_'):
            continue

        if select is None:
            sel = None
        elif '*' in select:
            sel = select['*']
        elif key in select:
            sel = select[key]
        else:
            continue

        if sel is not None and sel == {}:
            sel = {'*': {}}

        yield key, sel


# FIXME: We should check select list at the very beginning of
#        request, not when returning results.
def check_unknown_props(
    node: Union[Model, Property, DataType],
    select: Optional[Iterable[str]],
    known: Iterable[str],
):
    unknown_properties = set(select or []) - set(known) - {'*'}
    if unknown_properties:
        raise exceptions.MultipleErrors(
            exceptions.FieldNotInResource(node, property=prop)
            for prop in sorted(unknown_properties)
        )


def flat_select_to_nested(select: Optional[List[str]]) -> SelectTree:
    """
    >>> flat_select_to_nested(None)

    >>> flat_select_to_nested(['foo.bar'])
    {'foo': {'bar': {}}}

    """
    if select is None:
        return None

    res = {}
    for v in select:
        if isinstance(v, dict):
            v = spyna.unparse(v)
        names = v.split('.')
        vref = res
        for name in names:
            if name not in vref:
                vref[name] = {}
            vref = vref[name]

    return res


def get_model_reserved_props(action: Action) -> List[str]:
    if action in (Action.GETALL, Action.SEARCH):
        return ['_type', '_id', '_revision']
    elif action == Action.CHANGES:
        return ['_cid', '_created', '_op', '_id', '_txn', '_revision']
    else:
        return ['_type', '_id', '_revision']


def get_ns_reserved_props(action: Action) -> List[str]:
    return []



from typing import Any

from spinta.core.ufuncs import Bind
from spinta.core.ufuncs import Env
from spinta.core.ufuncs import ufunc


@ufunc.resolver(Env, Bind)
def param(env: Env, bind: Bind) -> Any:
    return env.params[bind.name]



from starlette.requests import Request

from spinta.accesslog import AccessLog
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.compat import urlparams_to_expr
from spinta import commands
from spinta.components import Context, Model, Property, Action, UrlParams
from spinta.renderer import render
from spinta.core.ufuncs import Expr
from spinta.types.datatype import DataType, File, Object
from spinta.exceptions import ItemDoesNotExist, UnavailableSubresource
from spinta.backends.mongo.components import Mongo
from spinta.backends.mongo.commands.query import MongoQueryBuilder


@commands.getone.register(Context, Request, Model, Mongo)
async def getone(
    context: Context,
    request: Request,
    model: Model,
    backend: Mongo,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)

    accesslog = context.get('accesslog')
    accesslog.log(
        model=model.model_type(),
        action=action.value,
        id_=params.pk,
    )

    data = getone(context, model, backend, id_=params.pk)
    select_tree = get_select_tree(context, action, params.select)
    prop_names = get_select_prop_names(
        context,
        model,
        model.properties,
        action,
        select_tree,
    )
    data = commands.prepare_data_for_response(
        context,
        model,
        params.fmt,
        data,
        action=action,
        select=select_tree,
        prop_names=prop_names,
    )
    return render(context, request, model, params, data, action=action)


@commands.getone.register(Context, Model, Mongo)
def getone(
    context: Context,
    model: Model,
    backend: Mongo,
    *,
    id_: str,
):
    table = backend.db[model.model_type()]
    keys = {k: 1 for k in model.flatprops}
    keys['__id'] = 1
    keys['_id'] = 0
    query = {'__id': id_}
    data = table.find_one(query, keys)
    if data is None:
        raise ItemDoesNotExist(model, id=id_)
    data['_id'] = data['__id']
    data['_type'] = model.model_type()
    return commands.cast_backend_to_python(context, model, backend, data)


@commands.getone.register(Context, Request, Property, DataType, Mongo)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: DataType,
    backend: Mongo,
    *,
    action: Action,
    params: UrlParams,
):
    raise UnavailableSubresource(prop=prop.name, prop_type=prop.dtype.name)


@commands.getone.register(Context, Request, Property, (Object, File), Mongo)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: (Object, File),
    backend: Mongo,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, prop)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.log(
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )

    data = getone(context, prop, dtype, backend, id_=params.pk)
    data = commands.prepare_data_for_response(
        context,
        prop.dtype,
        params.fmt,
        data,
        action=Action.GETONE,
    )
    return render(context, request, prop, params, data, action=action)


@commands.getone.register(Context, Property, Object, Mongo)
def getone(
    context: Context,
    prop: Property,
    dtype: Object,
    backend: Mongo,
    *,
    id_: str,
):
    table = backend.db[prop.model.model_type()]
    data = table.find_one({'__id': id_}, {
        '__id': 1,
        '_revision': 1,
        prop.name: 1,
    })
    if data is None:
        raise ItemDoesNotExist(prop, id=id_)
    result = {
        '_id': data['__id'],
        '_revision': data['_revision'],
        '_type': prop.model_type(),
        prop.name: (data.get(prop.name) or {}),
    }
    return commands.cast_backend_to_python(context, prop, backend, result)


@commands.getone.register(Context, Property, File, Mongo)
def getone(
    context: Context,
    prop: Property,
    dtype: File,
    backend: Mongo,
    *,
    id_: str,
):
    table = backend.db[prop.model.model_type()]
    data = table.find_one({'__id': id_}, {
        '__id': 1,
        '_revision': 1,
        prop.name: 1,
    })
    if data is None:
        raise ItemDoesNotExist(prop, id=id_)
    # merge file property data with defaults
    file_data = {
        **{
            '_content_type': None,
            '_id': None,
            '_size': None,
        },
        **data.get(prop.name, {}),
    }
    result = {
        '_id': data['__id'],
        '_revision': data['_revision'],
        '_type': prop.model_type(),
        prop.name: file_data,
    }
    return commands.cast_backend_to_python(context, prop, backend, result)


@commands.getall.register(Context, Model, Mongo)
def getall(
    context: Context,
    model: Model,
    backend: Mongo,
    *,
    query: Expr = None,
):
    builder = MongoQueryBuilder(context)
    builder.update(model=model)
    table = backend.db[model.model_type()]
    env = builder.init(backend, table)
    expr = env.resolve(query)
    where = env.execute(expr)
    cursor = env.build(where)
    for row in cursor:
        if '__id' in row:
            row['_id'] = row.pop('__id')
        row['_type'] = model.model_type()
        yield commands.cast_backend_to_python(context, model, backend, row)



import cgi
import pathlib
import typing

from starlette.exceptions import HTTPException
from starlette.requests import Request

from spinta import commands
from spinta.accesslog import AccessLog

from spinta.renderer import render
from spinta.utils.aiotools import aiter
from spinta.components import Context, Action, UrlParams, DataItem
from spinta.commands.write import prepare_patch, simple_response, validate_data
from spinta.types.datatype import File
from spinta.backends.fs.components import FileSystem

if typing.TYPE_CHECKING:
    from spinta.backends.postgresql.components import WriteTransaction


@commands.push.register()
async def push(
    context: Context,
    request: Request,
    dtype: File,
    backend: FileSystem,
    *,
    action: Action,
    params: UrlParams,
):
    prop = dtype.prop

    commands.authorize(context, action, prop)

    transaction: WriteTransaction = context.get('transaction')
    accesslog: AccessLog = context.get('accesslog')
    accesslog.log(
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
        rev=request.headers.get('revision'),
        txn=transaction.id,
    )

    data = DataItem(
        prop.model,
        prop,
        propref=False,
        backend=backend,
        action=action
    )
    data.given = {
        prop.name: {
            '_content_type': request.headers.get('content-type'),
            '_id': None,
        }
    }

    if 'revision' in request.headers:
        data.given['_revision'] = request.headers.get('revision')
    if 'content-disposition' in request.headers:
        data.given[prop.name]['_id'] = cgi.parse_header(request.headers['content-disposition'])[1]['filename']
    if 'content-length' not in request.headers:
        raise HTTPException(status_code=411)
    if not data.given[prop.name]['_id']:
        # XXX: Probably here should be a new UUID.
        data.given[prop.name]['_id'] = params.pk

    commands.simple_data_check(context, data, data.prop, data.model.backend)

    data.saved = commands.getone(context, prop, dtype, prop.model.backend, id_=params.pk)

    filepath = backend.path / data.given[prop.name]['_id']

    dstream = aiter([data])
    dstream = validate_data(context, dstream)
    dstream = prepare_patch(context, dstream)
    if action in (Action.UPDATE, Action.PATCH):
        dstream = create_file(filepath, dstream, request.stream())
        dstream = commands.update(context, prop, dtype, prop.model.backend, dstream=dstream)
    elif action == Action.DELETE:
        dstream = commands.delete(context, prop, dtype, prop.model.backend, dstream=dstream)
    else:
        raise Exception(f"Unknown action {action!r}.")
    dstream = commands.create_changelog_entry(
        context, prop.model, prop.model.backend, dstream=dstream,
    )

    status_code, response = await simple_response(context, params.fmt, dstream)
    return render(context, request, prop, params, response, action=action, status_code=status_code)


async def create_file(
    filepath: pathlib.PosixPath,
    dstream: typing.AsyncIterator[DataItem],
    fstream: typing.AsyncIterator[bytes]
) -> typing.AsyncGenerator[DataItem, None]:
    # we know that dstream contains one DataItem element
    async for d in dstream:
        with open(filepath, 'wb') as f:
            async for chunk in fstream:
                f.write(chunk)
        yield d



from spinta import commands
from spinta.components import Context
from spinta.backends.mongo.components import Mongo


@commands.bootstrap.register(Context, Mongo)
def bootstrap(context: Context, backend: Mongo):
    pass



import sqlalchemy as sa

from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import FunctionElement
from sqlalchemy.dialects.mssql.base import ischema_names
from geoalchemy2.types import Geometry


ischema_names['geometry'] = Geometry


class utcnow(FunctionElement):
    type = sa.DateTime()


@compiles(utcnow, 'postgresql')
def pg_utcnow(element, compiler, **kw):
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"



from __future__ import annotations

from typing import TYPE_CHECKING

import sqlalchemy as sa

from sqlalchemy.dialects.postgresql import JSONB, BIGINT

from spinta import commands
from spinta.components import Context, Model
from spinta.backends.postgresql.constants import TableType
from spinta.backends.postgresql.helpers import get_pg_name
from spinta.backends.postgresql.helpers import get_table_name

if TYPE_CHECKING:
    from spinta.backends.postgresql.components import PostgreSQL


def get_changes_table(context: Context, backend: PostgreSQL, model: Model):
    table_name = get_pg_name(get_table_name(model, TableType.CHANGELOG))
    pkey_type = commands.get_primary_key_type(context, backend)
    table = sa.Table(
        table_name, backend.schema,
        # XXX: This will not work with multi master setup. Consider changing it
        #      to UUID or something like that.
        #
        #      `change` should be monotonically incrementing, in order to
        #      have that, we could always create new `change_id`, by querying,
        #      previous `change_id` and increment it by one. This will create
        #      duplicates, but we simply know, that these changes happened at at
        #      the same time. So that's probably OK.
        sa.Column('_id', BIGINT, primary_key=True),
        sa.Column('_revision', sa.String),
        sa.Column('_txn', pkey_type, index=True),
        sa.Column('_rid', pkey_type),  # reference to main table
        sa.Column('datetime', sa.DateTime),
        # FIXME: Change `action` to `_op` for consistency.
        sa.Column('action', sa.String(8)),  # insert, update, delete
        sa.Column('data', JSONB),
    )
    return table



from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.constants import TableType
from spinta.backends.postgresql.helpers import get_column_name, get_table_name, \
    get_pg_name
from spinta.components import Context, Model
from spinta.types.datatype import DataType, Object, Array
from spinta.migrations import SchemaVersion
from spinta.utils.data import take


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, type(None), Model)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: type(None),
    current: Model,
):
    props = []
    for prop in take(['_id', '_revision', all], current.properties).values():
        props.extend(
            freeze(context, version, backend, freezed, prop.dtype)
        )
    version.actions.append({
        'type': 'schema',
        'upgrade': {
            'name': 'create_table',
            'args': [current.name] + props,
        },
        'downgrade': {
            'name': 'drop_table',
            'args': [current.name],
        },
    })


def dzip(*data):
    keys = set()
    for d in data:
        keys.update(d)
    for key in keys:
        yield tuple(d.get(key) for d in data)


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, Model, Model)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: Model,
    current: Model,
):
    for freezed_prop, current_prop in dzip(freezed.properties, current.properties):
        freeze(
            context,
            version,
            backend,
            freezed_prop.dtype if freezed_prop else freezed,
            current_prop.dtype if current_prop else current,
        )


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, type(None), DataType)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: type(None),
    current: DataType,
):
    column = {
        'name': 'column',
        'args': [
            current.prop.place,
            {
                'name': current.name,
                'args': [],
            }
        ]
    }
    if current.nullable:
        column['args'].append({'name': 'bind', 'args': ['nullable', True]})
    return [column]


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, Model, DataType)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: Model,
    current: DataType,
):
    column = {
        'name': 'column',
        'args': [
            get_column_name(current.prop),
            {'name': current.name, 'args': []}
        ]
    }
    if current.nullable:
        column['args'].append({'name': 'bind', 'args': ['nullable', True]})

    upgrade = {
        'name': 'add_column',
        'args': [
            get_table_name(current.prop),
            column,
        ]
    }

    version.actions.append({
        'type': 'schema',
        'upgrade': upgrade,
        'downgrade': {
            'name': 'drop_column',
            'args': [
                get_table_name(current.prop),
                get_column_name(current.prop),
            ]
        }
    })


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, DataType, Model)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: DataType,
    current: Model,
):
    pass


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, DataType, DataType)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: DataType,
    current: DataType,
):
    table_name = get_table_name(current.prop)
    if current.prop.list:
        table_name = get_pg_name(get_table_name(current.prop, TableType.LIST))
    if freezed.name != current.name:
        actions = {
            'type': 'schema',
            'upgrade': {
                'name': 'alter_column',
                'args': [
                    table_name,
                    get_column_name(current.prop),
                    {
                        'name': 'bind',
                        'args': [
                            'type_',
                            {'name': current.name, 'args': []}
                        ]
                    },
                ]
            },
            'downgrade': {
                'name': 'alter_column',
                'args': [
                    table_name,
                    get_column_name(current.prop),
                    {
                        'name': 'bind',
                        'args': [
                            'type_',
                            {'name': freezed.name, 'args': []}
                        ]
                    },
                ]
            }
        }
        if freezed.nullable != current.nullable:
            actions['upgrade']['args'].append(
                {'name': 'bind', 'args': ['nullable', current.nullable]}
            )
            actions['downgrade']['args'].append(
                {'name': 'bind', 'args': ['nullable', freezed.nullable]}
            )
        version.actions.append(actions)
    elif freezed.nullable != current.nullable:
        version.actions.append({
            'type': 'schema',
            'upgrade': {
                'name': 'alter_column',
                'args': [
                    table_name,
                    get_column_name(current.prop),
                    {'name': 'bind', 'args': ['nullable', current.nullable]},
                ]
            },
            'downgrade': {
                'name': 'alter_column',
                'args': [
                    table_name,
                    get_column_name(current.prop),
                    {'name': 'bind', 'args': ['nullable', freezed.nullable]},
                ]
            }
        })



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set
from typing import TypedDict

import frictionless
from sqlalchemy.engine.base import Engine as SaEngine

from spinta import commands
from spinta import spyna
from spinta.components import Context
from spinta.components import Model
from spinta.components import Property
from spinta.datasets.backends.sql.components import Sql
from spinta.datasets.backends.sql.ufuncs.components import Engine
from spinta.datasets.backends.sql.ufuncs.components import SqlResource
from spinta.datasets.backends.sql.frictionless import GeoSqlStorage
from spinta.datasets.components import Resource
from spinta.exceptions import UnexpectedFormulaResult
from spinta.manifests.components import ManifestSchema
from spinta.manifests.components import NodeSchema
from spinta.manifests.helpers import entity_to_schema
from spinta.manifests.helpers import model_to_schema
from spinta.utils.imports import full_class_name
from spinta.utils.naming import Deduplicator
from spinta.utils.naming import to_model_name
from spinta.utils.naming import to_property_name


def _ensure_list(value: Optional[Any]):
    if value is None:
        return value
    elif isinstance(value, list):
        return value
    else:
        return [value]


class _FrictionlessForeignKeyReference(TypedDict):
    fields: List[str]
    resource: str


class _FrictionlessForeignKey(TypedDict):
    fields: List[str]
    reference: _FrictionlessForeignKeyReference


@dataclass()
class _ForeignKey:
    resource: frictionless.Resource
    field: frictionless.Field
    foreign_key: _FrictionlessForeignKey


@dataclass()
class _CompositeForeignKey:
    resource: frictionless.Resource
    foreign_key: _FrictionlessForeignKey


_ForeignKeysByProp = Dict[str, _FrictionlessForeignKey]
_ForeignKeysByField = Dict[str, _FrictionlessForeignKey]


def _read_frictionless_field(field: frictionless.Field) -> Dict[str, Any]:
    return {
        'type': field.type,
        'external': {
            'name': field.name,
        }
    }


@commands.inspect.register(Context, Sql, Model, _ForeignKey)
def inspect(
    context: Context,
    backend: Sql,
    model: Model,
    source: _ForeignKey,
) -> NodeSchema:
    ref_model_name = to_model_name(source.foreign_key['reference']['resource'])
    if ref_model_name:
        if model.external:
            ref_model_name = f'{model.external.dataset.name}/{ref_model_name}'
    else:
        # If ref_model_name is empty string, it means, this is a self reference.
        ref_model_name = model.name

    return {
        'type': 'ref',
        'model': ref_model_name,
        'refprops': [
            to_property_name(ref_field)
            for ref_field in source.foreign_key['reference']['fields']
        ],
        'external': {
            'name': source.field.name,
        }
    }


@commands.inspect.register(Context, Sql, Model, _CompositeForeignKey)
def inspect(
    context: Context,
    backend: Sql,
    model: Model,
    source: _CompositeForeignKey,
) -> NodeSchema:
    ref_model_name = to_model_name(source.foreign_key['reference']['resource'])
    if ref_model_name:
        if model.external:
            ref_model_name = f'{model.external.dataset.name}/{ref_model_name}'
    else:
        # If ref_model_name is empty string, it means, this is a self reference.
        ref_model_name = model.name

    return {
        'type': 'ref',
        'model': ref_model_name,
        'refprops': [
            to_property_name(ref_field)
            for ref_field in source.foreign_key['reference']['fields']
        ],
    }


@commands.inspect.register(Context, Sql, Resource, _ForeignKey)
def inspect(
    context: Context,
    backend: Sql,
    resource: Resource,
    source: _ForeignKey,
) -> NodeSchema:
    ref_model_name = source.foreign_key['reference']['resource']
    if ref_model_name:
        ref_model_name = to_model_name(ref_model_name)
    else:
        # If ref_model_name is empty string, it means, this is a self reference.
        ref_model_name = to_model_name(source.resource.name)
    ref_model_name = f'{resource.dataset.name}/{ref_model_name}'
    return {
        'type': 'ref',
        'model': ref_model_name,
        'refprops': [
            to_property_name(ref_field)
            for ref_field in source.foreign_key['reference']['fields']
        ],
        'external': {
            'name': source.field.name,
        }
    }


@commands.inspect.register(Context, Sql, Property, _ForeignKey)
def inspect(
    context: Context,
    backend: Sql,
    prop: Property,
    source: _ForeignKey,
) -> NodeSchema:
    ref_model_name = source.foreign_key['reference']['resource']
    if ref_model_name:
        ref_model_name = to_model_name(ref_model_name)
    else:
        # If ref_model_name is empty string, it means, this is a self reference.
        ref_model_name = to_model_name(source.resource.name)
    ref_model_name = f'{prop.model.external.dataset.name}/{ref_model_name}'
    return {
        'type': 'ref',
        'model': ref_model_name,
        'refprops': [
            to_property_name(ref_field)
            for ref_field in source.foreign_key['reference']['fields']
        ],
        'external': {
            'name': source.field.name,
            'prepare': prop.external.prepare,
        },
        'access': prop.given.access,
        'title': prop.title,
        'description': prop.description,
    }


@commands.inspect.register(Context, Sql, Property, frictionless.Field)
def inspect(
    context: Context,
    backend: Sql,
    prop: Property,
    source: frictionless.Field,
) -> NodeSchema:
    return {
        'type': source.type,
        'external': {
            'name': source.name,
            'prepare': prop.external.prepare,
        },
        'access': prop.given.access,
        'title': prop.title,
        'description': prop.description,
    }


@commands.inspect.register(Context, Sql, Model, frictionless.Field)
def inspect(
    context: Context,
    backend: Sql,
    model: Model,
    source: frictionless.Field,
) -> NodeSchema:
    return {
        'type': source.type,
        'external': {
            'name': source.name,
        }
    }


@commands.inspect.register(Context, Sql, Resource, frictionless.Field)
def inspect(
    context: Context,
    backend: Sql,
    resource: Resource,
    source: frictionless.Field,
) -> NodeSchema:
    return {
        'type': source.type,
        'external': {
            'name': source.name,
        }
    }


@commands.inspect.register(Context, Sql, Property, type(None))
def inspect(
    context: Context,
    backend: Sql,
    prop: Property,
    source: frictionless.Field,
) -> NodeSchema:
    return {
        'type': prop.dtype.name,
        'external': {
            'name': prop.name,
            'prepare': prop.external.prepare,
        }
    }


@commands.inspect.register(Context, Sql, Model, frictionless.Resource)
def inspect(
    context: Context,
    backend: Sql,
    model: Model,
    source: frictionless.Resource,
) -> Iterator[ManifestSchema]:
    schema = source.schema
    eid, data = model_to_schema(model)

    if model.external:
        data['external'] = entity_to_schema(model.external)[1]
        if not data['external']['pk']:
            data['external']['pk'] = [
                to_property_name(p)
                for p in _ensure_list(schema.primary_key)
            ]
    else:
        data['external'] = {}

    data['external']['name'] = source.name

    source_props: Set[str] = set()
    model_props = {
        prop.external.name: prop
        for prop in model.properties.values()
        if prop.external
    }

    foreign_keys = _get_foreign_keys_by_field(schema.foreign_keys)

    props = {}
    deduplicate = Deduplicator('_{}')
    for field in schema.fields:
        source_props.add(field.name)

        if field.name in model_props:
            prop = model_props[field.name]
            node = prop
            name = prop.name
            deduplicate(name)
        else:
            node = model
            name = to_property_name(field.name)
            name = deduplicate(name)

        if field.name in foreign_keys:
            fk = _ForeignKey(source, field, foreign_keys[field.name])
            props[name] = commands.inspect(context, backend, node, fk)
        else:
            props[name] = commands.inspect(context, backend, node, field)

    for field, prop in model_props.items():
        if field not in source_props:
            props[prop.name] = commands.inspect(context, backend, prop, None)

    for field, fk in foreign_keys.items():
        if field not in source_props:
            cfk = _CompositeForeignKey(source, fk)
            props[field] = commands.inspect(context, backend, model, cfk)

    data['properties'] = props

    yield eid, data


def _get_foreign_keys_by_field(
    foreign_keys: List[_FrictionlessForeignKey],
) -> _ForeignKeysByField:
    result: _ForeignKeysByField = {}
    for fk in foreign_keys:
        name = '_'.join(fk['fields'])
        result[name] = fk
    return result


@commands.inspect.register(Context, Sql, Resource, frictionless.Resource)
def inspect(
    context: Context,
    backend: Sql,
    resource: Resource,
    source: frictionless.Resource,
) -> Iterator[ManifestSchema]:
    """Add a new model"""
    schema = source.schema
    eid, data = model_to_schema(None)

    deduplicate: Deduplicator = context.get('deduplicate.model')
    data['name'] = deduplicate(to_model_name(source.name))
    data['external'] = {
        'name': source.name,
        'pk': [
            to_property_name(p)
            for p in _ensure_list(schema.primary_key)
        ]
    }

    foreign_keys = _get_foreign_keys_by_field(schema.foreign_keys)

    props = {}
    deduplicate = Deduplicator('_{}')
    for field in schema.fields:
        name = deduplicate(to_property_name(field.name))
        if field.name in foreign_keys:
            fk = _ForeignKey(source, field, foreign_keys[field.name])
            props[name] = commands.inspect(context, backend, resource, fk)
        else:
            props[name] = commands.inspect(context, backend, resource, field)

    data['properties'] = props

    yield eid, data


@commands.inspect.register(Context, Sql, Model, type(None))
def inspect(
    context: Context,
    backend: Sql,
    model: Model,
    source: None,
) -> Iterator[ManifestSchema]:
    """Keep an existing model that is no longer in the resource"""
    yield model_to_schema(model)


@commands.inspect.register(Context, Sql, Resource, type(None))
def inspect(
    context: Context,
    backend: Sql,
    resource: Resource,
    source: None,
) -> Iterator[ManifestSchema]:
    if resource.prepare:
        env = SqlResource(context).init(backend.config['dsn'])
        engine = env.resolve(resource.prepare)
        engine: Engine = env.execute(engine)
        if not isinstance(engine, Engine):
            raise UnexpectedFormulaResult(
                resource,
                formula=spyna.unparse(resource.prepare),
                expected=full_class_name(SaEngine),
                received=full_class_name(engine),
            )
    else:
        engine = Engine(backend.config['dsn'])

    resource_models: Set[str] = set()
    manifest_models = {
        model.external.name: model
        for model in resource.models.values()
        if model.external and model.external.name
    }

    storage = GeoSqlStorage(
        engine=engine.create(),
        namespace=engine.schema,
    )
    package = frictionless.Package.from_storage(storage)
    with context:
        context.bind('deduplicate.model', Deduplicator)

        for i, source_ in enumerate(package.resources):
            model = manifest_models.get(source_.name)
            resource_models.add(source_.name)
            schemas = commands.inspect(context, backend, model or resource, source_)
            for eid, schema in schemas:
                if 'external' not in schema:
                    schema['external'] = {}
                schema['external']['dataset'] = resource.dataset.name
                schema['external']['resource'] = resource.name
                if model is None:
                    schema['name'] = resource.dataset.name + '/' + schema['name']
                yield eid, schema

        for source_, model in manifest_models.items():
            if source_ not in resource_models:
                for eid, schema in commands.inspect(context, backend, model, None):
                    if 'external' not in schema:
                        schema['external'] = {}
                    schema['external']['dataset'] = resource.dataset.name
                    schema['external']['resource'] = resource.name
                    yield eid, schema



from __future__ import annotations

from typing import TYPE_CHECKING

from spinta.core.ufuncs import asttoexpr

if TYPE_CHECKING:
    from spinta.components import UrlParams
    from spinta.core.ufuncs import Expr


def urlparams_to_expr(
    params: UrlParams,
    # XXX: `add_count` is a hack, because, not all backends supports it yet.
    add_count: bool = True,
) -> Expr:
    """Convert UrlParams to ufunc Expr.

    This is compatibility function. Currently all query handling is done using
    UrlParams, but in future UrlParams should be replaced by ufunc Expr. This
    way getall and getone commands will receive just one Expr parameter
    containing all the query information.
    """

    ast = []

    if params.query:
        ast += params.query

    if params.select:
        ast.append({'name': 'select', 'args': [
            arg if isinstance(arg, dict) else {'name': 'bind', 'args': [arg]}
            for arg in params.select
        ]})

    if params.sort:
        ast.append({'name': 'sort', 'args': [
            arg if isinstance(arg, dict) else {'name': 'bind', 'args': [arg]}
            for arg in params.sort
        ]})

    if params.limit:
        ast.append({'name': 'limit', 'args': [params.limit]})

    if params.offset:
        ast.append({'name': 'offset', 'args': [params.offset]})

    if params.count and add_count:
        ast.append({'name': 'count', 'args': []})

    if len(ast) == 0:
        ast = {
            'name': 'select',
            'args': [],
        }
    elif len(ast) == 1:
        ast = ast[0]
    else:
        ast = {'name': 'and', 'args': ast}

    return asttoexpr(ast)



from typing import TextIO, Union

import json
import sys
import pathlib

from starlette.requests import Request

from spinta import commands
from spinta.accesslog import AccessLog
from spinta.components import Context, Config, Store


class FileAccessLog(AccessLog):
    file: Union[TextIO, pathlib.Path] = None

    _file: TextIO = None
    _closed: bool = True

    def __exit__(self, *exc):
        if not self._closed:
            self._file.close()

    def log(self, **kwargs):
        if self._file is None:
            return
        message = self.create_message(**kwargs)
        message = json.dumps(message)
        print(message, file=self._file)


@commands.load.register(Context, FileAccessLog, Config)
def load(context: Context, accesslog: FileAccessLog, config: Config):
    commands.load[Context, AccessLog, Config](context, accesslog, config)

    file = config.rc.get('accesslog', 'file')

    if file == 'stdout':
        file = sys.stdout
    elif file == 'stderr':
        file = sys.stderr
    elif file in ('null', '/dev/null'):
        file = None
    else:
        file = pathlib.Path(file)

        # Try to open file to check if it is writable.
        with file.open('a'):
            pass

    accesslog.file = file
    return accesslog


def _prepare_file(
    accesslog: AccessLog,
    file: Union[TextIO, pathlib.Path, None],
) -> None:
    if isinstance(file, pathlib.Path):
        file = file.open('a')
        accesslog._closed = False
    else:
        accesslog._closed = True
    accesslog._file = file


@commands.load.register(Context, FileAccessLog, Store)
def load(context: Context, accesslog: FileAccessLog, store: Store):
    if isinstance(store.accesslog, FileAccessLog):
        _prepare_file(accesslog, store.accesslog.file)


@commands.load.register(Context, FileAccessLog, Request)
def load(context: Context, accesslog: FileAccessLog, request: Request):
    commands.load[Context, AccessLog, Request](context, accesslog, request)
    store: Store = context.get('store')
    if isinstance(store.accesslog, FileAccessLog):
        _prepare_file(accesslog, store.accesslog.file)



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union
from typing import overload

from starlette.requests import Request

from spinta.components import UrlParams
from spinta.components import Version
from spinta.dispatcher import command
from spinta.manifests.components import ManifestSchema
from spinta.manifests.components import NodeSchema

if TYPE_CHECKING:
    from spinta.components import Store
    from spinta.components import Model
    from spinta.components import Property
    from spinta.types.datatype import DataType
    from spinta.components import Action
    from spinta.backends import Backend
    from spinta.components import Context
    from spinta.manifests.components import Manifest
    from spinta.datasets.components import Dataset
    from spinta.datasets.components import Resource
    from spinta.formats.components import Format
    from spinta.formats.html.components import ComplexCell


T = TypeVar('T')


@command()
def manifest_list_schemas():
    """Return iterator of manifest schema entry ids.

    For YAML manifests entry id is file path, for backend manifests, entry id is
    _id, for tabular manifests, entry id is table row id.

    """


@command()
def manifest_read_current():
    """Return current schema by given schema entry id.

    For YAML manifests, current schema is first YAML file document entry. For
    Backend manifests, current schema is read from _schema table.
    """


@command()
def manifest_read_freezed():
    """Return last freezed schema by given schema entry id.

    Freezed schema is schema of last freezed version.
    """


@command()
def manifest_read_versions():
    """Return iterator of all schema versions by given schema entry id."""


@command()
def configure():
    """Configure component before loading.

    This is a very first thing that happens to a component.
    """


@command()
def load(
    context: Context,
    manifest: Manifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
    # Do not raise error, when trying to load a node, with a name, that is
    # already loaded, instead rename the new node.
    rename_duplicates: bool = False,
    # If True, load internal manifest, if not loaded.
    load_internal: bool = True,
) -> None:
    """Load primitive data structures to python-native objects.

    Currently used for:

    - Load things from configuration:

        load(Context, Config, RawConfig) -> Config
        load(Context, Store, Config) -> Store
        load(Context, Backend, Config) -> Backend
        load(Context, Manifest, dict) -> Manifest

    - Load nodes from manifest:

        load(Context, Node, dict, Manifest) -> Node

    - Load commands from manifest:

        load(Context, Command, dict, *, scope=None) -> Command

    - Load primitive data types to python-native objects:

        load(Context, X, Node, Backend) -> Y

    - Load data from request:

        load(Context, Property, Request)

    """


@command()
def decode():
    """Decode given value from source backend format into target backend format.

    decode(Context, SourceFormat, DataType, value)
    decode(Context, SourceBackend, DataType, value)
    decode(Context, SourceFormat, TargetBackend, DataType, value)

    """


@command()
def link():
    """Link loaded components.

    While loading, components can't be linked, because not all components might
    be loaded. When linking, all components are loaded and can be linked.
    """
    pass


@command()
def load_search_params():
    """Load search parameters as native python values.
    """


@command()
def load_operator_value():
    """Loads URL query operator value given by the user.
    """


@overload
def wait(
    context: Context,
    store: Store,
    *,
    seconds: int = None,
    verbose: bool = False,
) -> bool:
    """Wait for all backends given amount of seconds"""


@overload
def wait(context: Context, backend: Backend, *, fail: bool = False) -> bool:
    """Check if backend is available and operational

    This usually attempts to connect to backend with a very small timeout and
    return True if connection is successful or False if not.

    If `fail` is False, catch all connection related errors and return False,
    if True, raise exception as is.
    """


@command()
def wait(*args):
    """Wait while all database backends are up.

    Database backends are a separate services and once you start multiple
    services at the same time, you need to make sure, that all the required
    external services are up before running the app.
    """


@command()
def check():
    """Check if input value is correct.

    - Check components after load:

        check(Context, Config):
        check(Context, Store):
        check(Context, Manifest):
        check(Context, Node):

    - Check data before push:

        check(Context, Model, Backend, dict, *, action: Action)
        check(Context, Type, Property, Backend, object, *, data: dict, action: Action)

    """


@command()
def simple_data_check():
    """Run a simple data check.

    Check data by using restrictions given in schema, do not run any business
    logic. Also simple data check should not do any database queries.

    Simple data check is run early, before retrieving exisisting data from
    database.

    Later a complex data check is run with existing data loaded from database.
    """


@command()
def complex_data_check():
    """Run a complex data check.

    At this point, simple data check is already done and data is passed simple
    validation.

    Complex data check will receive existing data in database, additional
    queries can be run to do more complex validations.
    """


@command()
def check_unique_constraint():
    """Check if value is unique.

    This check is only performed when peroperty has unique property set to true.
    """


@overload
def prepare(
    context: Context,
    backend: Backend,
    dtype: Manifest,
) -> None:
    """Prepare whole manifest

    Basically this initializes database schema objects, with sqlalchemy it
    constructs whole metadata schema, like tables and columns.

    This command should call prepare[Context, Backend, Property].
    """


@overload
def prepare(
    context: Context,
    backend: Backend,
    dtype: Model,
) -> None:
    """Prepare model

    This command should call prepare[Context, Backend, Property].
    """


@overload
def prepare(
    context: Context,
    backend: Backend,
    dtype: Property,
) -> Union[List[T], T, None]:
    """Prepare model property

    This command should call prepare[Context, Backend, DataType].

    Returns list of columns, single column or None.
    """


@overload
def prepare(
    context: Context,
    backend: Backend,
    dtype: DataType,
) -> Union[List[T], T, None]:
    """Prepare model property data type

    Returns list of columns, single column or None.
    """


@overload
def prepare(
    context: Context,
    params: UrlParams,
    version: Version,
    request: Request,
) -> UrlParams:
    """Prepare UrlParams from a Request."""


@command()
def prepare(*args, **kwargs):
    """Prepare value.

    - Prepare database backend:

        prepare(Context, Store)
        prepare(Context, Manifest)
        prepare(Context, Backend, Manifest)
        prepare(Context, Backend, Node)
        prepare(Context, Backend, Type)

      Here, sqlalchemy.MetaData object is populated with tables and columns.

    - Build UrlParams from Request:

        prepare(Context, UrlParams, Version, Request) -> UrlParams

    - Prepare external dataset for data import:

        prepare(Context, Source, Node)

    """


@overload
def prepare_for_write(
    context: Context,
    model: Model,
    backend: Backend,
    patch_: Dict[str, Any],
    action: Action,
):
    """Convert Python-native Model patch data to backend-native patch"""


@overload
def prepare_for_write(
    context: Context,
    prop: Property,
    backend: Backend,
    patch_: Any,
):
    """Convert Python-native Property patch data to backend-native patch

    This is called, when writing data directly into property, for example:

        PATCH /model/:one/UUID/prop

    """


@overload
def prepare_for_write(
    context: Context,
    dtype: DataType,
    backend: Backend,
    value: Any,
):
    """Convert a Python-native value to a backend-native value for a specific
    data type

    This is usually called from another prepare_for_write command, from Model or
    a Property. But also can be called recursively, for nested data types, liek
    Object or Array.
    """


@command()
def prepare_for_write(
    context: Context,
    dtype: Union[Model, Property, DataType],
    backend: Backend,
    patch_: Any,
    **kwargs,
):
    """Convert Python-native values to backend-native values

        prepare(Context, Model, Backend, dict) -> dict
        prepare(Context, Property, Backend, object) -> object
        prepare(Context, Type, Backend, object) -> object

    """


@command()
def prepare_data_for_response():
    """Prepare Python-native data types for response.

    Prepared data must be json-serializable.

    This command is responsible for preparing model or property data at the top
    level of data structure. While prepare_dtype_for_response is responsible for
    preparing individual properties.
    """


@command()
def prepare_dtype_for_response():
    """Prepare Python-native data types for response.

    Prepared data must be json-serializable.

    This command is responsible for preparing individual properties for
    response. For the top level data preparation see prepare_data_for_response.
    """


@command()
def freeze():
    """Create new schema version.

    Freeze commands receive empty manifest instance, returned by create_manifest
    helper funciton. Then freeze command should fully load two versions of
    manijest, one version is current and another is freezed.

    Then these two manifests will be compared and a new version will be produced
    for each model if current differs from freezed.

    """


@command()
def bootstrap():
    """Bootstrap all components.

    Currently if database is empty instead of running all migrations, it creates
    all the tables and constrains directly.
    """


@command()
def sync():
    """Sync manifest YAML files to database."""


@command()
def migrate():
    """Migrate database schema changes."""


@command()
def authorize():
    """Check if user is authorized to access a resource."""


@command()
def push():
    """Insert, update or delete data to the databse."""


@command()
def insert():
    pass


@command()
def insert_many():
    pass


@command()
def upsert():
    pass


@command()
def upsert_many():
    pass


@command()
def update():
    pass


@command()
def update_many():
    pass


@command()
def patch():
    pass


@command()
def patch_many():
    pass


@command()
def delete():
    pass


@command()
def delete_many():
    pass


@command()
def getone():
    """Get single record from the databse."""


@command()
def getfile():
    """Get full file content

    This can raise error if file is to large to read into memeory.

    This command is used for file property, to read while file content into
    `_content` property if explicitly requested. For small files this is OK, but
    for larger files direct file access via subresource API should be used.
    """


@command()
def getall():
    """Find multiple records in the databse."""


@command()
def get_version():
    """Returns version dict for the api"""


@command()
def changes():
    """Changelog of a table."""


@command()
def pull():
    """Pull data from external data sources."""


@command()
def export():
    """Export data in a specified format."""


@command()
def wipe():
    """Delete all data from specified model."""


@command()
def get_primary_key_type():
    """Return primary key column type.

    This applyies to some backends, for example PostgreSQL, but is not used by
    other backends, like Mongo.

    PostgreSQL returns sqlalchemy.dialects.postgresql.UUID.
    """


@command()
def gen_object_id():
    """Genearet unique id.

    gen_object_id(Context, Backend, Model) -> str

    """


@command()
def is_object_id():
    """Check if given string is object id.

    - Detect if given string is an object id:

        is_object_id(Context, str) -> bool

    - Detect if given string is an object id of given backend and model.

        is_object_id(Context, Backend, Model, str) -> bool

    """


@command()
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Format,
    *,
    action: Action,
    params: UrlParams,
    data: Iterator[ComplexCell],
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    """Render response to the clinet."""


@command()
def unload_backend():
    """Unload database backend.

    Currently this is only used in tests and only for PostgreSQL, after each
    tests, all sqlalchemy connections are disposed to avoid leaving open
    connections between tests.
    """


@command()
def make_json_serializable():
    """Convert given Python-native value to a JSON serializable value."""


@command()
def get_error_context():
    """Get error context for a given object."""


@command()
def in_namespace():
    """Check if `a` is in `b` namespace."""


@command()
def create_changelog_entry():
    """Create new changelog entry."""


@command()
def coerce_source_value():
    """Coerce value received from external data source to python native type."""


@command()
def new_schema_version():
    """Calculate diff between two nodes."""


@command()
def build_data_patch_for_write():
    """Builds data patch dict for backend consumption on write.

    Purpose of this command is to generate a patch, by comparing given and saved
    values. If given and saved values are same, then we exclude this property
    from patch, because nothing has changed.

    Optionally patch can be filled with default values. That means, if given
    value does not have all properties defined in schema, then we fill those
    missing values with defaults and only then proceed with patch generation.

    In the end, this command produces partial patch, with only those values
    which has been changed.

    """


@command()
def before_write():
    """Prepare patch for backend before writing

    This command converts Python-native data types to backend-native data types
    and prepares patch that is ready for writing to database tables.

    Preparation also might involve cleaning stale data on other tables, for
    example list data are written to separate tables used for searches, file
    properties writes file blockes to a separate tables and updates patch with
    block data.
    """


@command()
def after_write():
    """Run additional actions after writing to backend

    This can invorvle writing data to other tables, for examples list can save
    list items, files can save file blocks, etc.

    At this point, main resource data are saved and all referential checks will
    pass.
    """


@command()
def build_full_response():
    """Builds full patch from data.patch and data.saved for a response

    This solves an issue, when we data.patch is missing some data from
    data.given because that data is already in data.saved, thus response
    instead of having values from data.saved - returns default values
    (usually `None` or empty strings), which is confusing to the user.
    """


@command()
def rename_metadata():
    """Renames metadata fields to be compatible with spinta's notation.

    XXX: this is actually a hack and should be refactored into something smarter
    """


@command()
def get_model_scopes():
    """Returns list of model scopes by given list of actions.
    """


@command()
def cast_backend_to_python():
    """Convert backend native types to python native types."""


@overload
def inspect(
    context: Context,
    backend: Backend,
    manifest: Optional[Manifest],
    source: Optional[Any],
) -> Iterator[ManifestSchema]:
    """Inspect manifest schema."""


@overload
def inspect(
    context: Context,
    backend: Backend,
    dataset: Optional[Dataset],
    source: Optional[Any],
) -> Iterator[ManifestSchema]:
    """Inspect dataset schema."""


@overload
def inspect(
    context: Context,
    backend: Backend,
    resource: Optional[Resource],
    source: Optional[Any],
) -> Iterator[ManifestSchema]:
    """Inspect dataset resource schema."""


@overload
def inspect(
    context: Context,
    backend: Backend,
    model: Optional[Model],
    source: Optional[Any],
) -> Iterator[ManifestSchema]:
    """Inspect model schema."""


@overload
def inspect(
    context: Context,
    backend: Backend,
    prop: Optional[Property],
    source: Optional[Any],
) -> NodeSchema:
    """Inspect model property schema."""


@overload
def inspect(
    context: Context,
    backend: Backend,
    dtype: Optional[DataType],
    source: Optional[Any],
) -> NodeSchema:
    """Inspect property data type schema."""


@command()
def inspect(*args) -> Iterator[ManifestSchema]:
    """Inspect schema of a node on a given backend.

    Usually manifest is loaded from a manifest backend, but inspect reads
    manifest metadata directly from storage backend. This way, one can generate
    manifest from an existing data source.

    Which data sources must be read will be discovered in the manifest. Inspect
    will read given manifest and then will rewrite schemas from backend.

    This is used in inspect command.

    All inspect commands takes backend and an existing node if a node is already
    in manifest. If node is not in manifest, then None will be given.

    Inspect commands must generate ManifestSchema objects for all inspected
    schema elements from a given backend, but also must return existing nodes
    even if they do not exist in the given backend.

    Nodes that exists in manifest, but not in backend, should clear source
    (external) values, indicating, that node is not present in the backend.

    If node exists on both, backend and manifest, then two nodes must be merged
    into one.

    When merging, backend should overwrite everything related to backend,
    leaving everything else as was in manifest.

    Existing manifest nodes must be matched with backend equivalents by
    comparing source names.

    """



from spinta import commands
from spinta.migrations import get_new_schema_version
from spinta.components import Context
from spinta.manifests.yaml.components import YamlManifest
from spinta.manifests.helpers import create_manifest
from spinta.manifests.helpers import get_current_schema_changes
from spinta.manifests.yaml.helpers import add_new_version


@commands.freeze.register(Context, YamlManifest)
def freeze(context: Context, current: YamlManifest):
    # Load current node, that is not yet freezed.
    commands.load(context, current, freezed=False)
    commands.link(context, current)
    commands.check(context, current)

    # Load previously freezed version, we don't need to run check, because
    # freezed version was already checked before freezing it.
    freezed = create_manifest(context, current.store, current.name)
    commands.load(context, freezed, freezed=True)
    commands.link(context, freezed)

    # Read current nodes, get freezed versions and freeze changes between
    # current and previously freezed node into a new version.
    for ntype, nodes in current.objects.items():

        if ntype == 'ns':
            # Namespaces do not have physical form, yet, se there is no need to
            # freeze them.
            continue

        for name, cnode in nodes.items():

            # Get freezed node
            if name in freezed.objects[ntype]:
                fnode = freezed.objects[ntype][name]
            else:
                fnode = None

            # Check if current and freezed schema versions changed
            changes = get_current_schema_changes(context, current, cnode.eid)
            if not changes:
                continue

            # Autodectect and write changes into version instance.
            version = get_new_schema_version(changes)
            commands.freeze(context, version, cnode.backend, fnode, cnode)

            # Write version data back to node file.
            print(f"Adding new version {version.id} to {cnode.eid}")
            add_new_version(cnode.eid, version)



from typing import Iterator

from multipledispatch import dispatch

from starlette.requests import Request

from spinta import commands
from spinta.accesslog import AccessLog
from spinta.renderer import render
from spinta.components import Context, Model, Property, Action, UrlParams
from spinta.types.datatype import DataType, File, Object
from spinta.exceptions import NotFoundError, ItemDoesNotExist
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name
from spinta.utils.nestedstruct import flat_dicts_to_nested


@commands.getone.register(Context, Request, Property, Object, PostgreSQL)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: Object,
    backend: PostgreSQL,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, prop)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.log(
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )

    data = getone(context, prop, dtype, backend, id_=params.pk)
    data = commands.prepare_data_for_response(
        context,
        prop.dtype,
        params.fmt,
        data,
        action=Action.GETONE,
    )
    return render(context, request, prop, params, data, action=action)


@commands.getone.register(Context, Property, Object, PostgreSQL)
def getone(
    context: Context,
    prop: Property,
    dtype: Object,
    backend: PostgreSQL,
    *,
    id_: str,
):
    table = backend.get_table(prop.model)
    connection = context.get('transaction').connection
    selectlist = [
        table.c._id,
        table.c._revision,
    ] + [
        table.c[name]
        for name in _iter_prop_names(prop.dtype)
    ]
    try:
        data = backend.get(connection, selectlist, table.c._id == id_)
    except NotFoundError:
        raise ItemDoesNotExist(prop.model, id=id_)

    result = {
        '_type': prop.model_type(),
        '_id': data[table.c._id],
        '_revision': data[table.c._revision],
    }

    data = flat_dicts_to_nested(data)
    result[prop.name] = data[prop.name]
    return commands.cast_backend_to_python(context, prop, backend, result)


@dispatch((Model, Object))
def _iter_prop_names(dtype) -> Iterator[Property]:
    for prop in dtype.properties.values():
        yield from _iter_prop_names(prop.dtype)


@dispatch(DataType)
def _iter_prop_names(dtype) -> Iterator[Property]:  # noqa
    if not dtype.prop.name.startswith('_'):
        yield get_column_name(dtype.prop)


@dispatch(File)
def _iter_prop_names(dtype) -> Iterator[Property]:  # noqa
    yield dtype.prop.place + '._id'
    yield dtype.prop.place + '._content_type'



from typing import Dict
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import cast

from spinta.components import Context
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Property
from spinta.core.access import link_access_param
from spinta.core.access import load_access_param
from spinta.core.ufuncs import asttoexpr
from spinta.dimensions.enum.components import EnumFormula
from spinta.dimensions.enum.components import EnumItem
from spinta.dimensions.enum.components import EnumValue
from spinta.dimensions.enum.components import Enums
from spinta.exceptions import ValueNotInEnum
from spinta.manifests.components import Manifest
from spinta.manifests.tabular.components import EnumRow
from spinta.nodes import load_node
from spinta.utils.schema import NA


def _load_enum_item(
    context: Context,
    parents: List[Union[Property, Model, Namespace, Manifest]],
    data: EnumRow,
) -> EnumItem:
    item = EnumItem()
    parent = parents[0]
    item = load_node(context, item, data, parent=parent)
    item = cast(EnumItem, item)
    if item.prepare is not NA:
        ast = item.prepare
        expr = asttoexpr(ast)
        env = EnumFormula(context, scope={
            'this': item.source,
            'node': parent,
        })
        item.prepare = env.resolve(expr)

    load_access_param(item, data.get('access'), parents)
    return item


def load_enums(
    context: Context,
    parents: List[Union[Property, Model, Namespace, Manifest]],
    enums: Optional[Dict[str, Dict[str, EnumRow]]],
) -> Optional[Enums]:
    if enums is None:
        return
    return {
        name: {
            source: _load_enum_item(context, parents, item)
            for source, item in enum.items()
        }
        for name, enum in enums.items()
    }


def link_enums(
    parents: List[Union[Property, Model, Namespace]],
    enums: Optional[Enums],
) -> None:
    if enums is None:
        return
    for enum in enums.values():
        for item in enum.values():
            link_access_param(item, parents)


def get_prop_enum(prop: Optional[Property]) -> EnumValue:
    if prop and prop.enum:
        return prop.enum


T = TypeVar('T')


def prepare_enum_value(prop: Property, value: T) -> Union[T, List[T]]:
    if enum := get_prop_enum(prop):
        source = [
            item.source
            for item in enum.values()
            if (
                (item.prepare is None and value is None) or
                item.prepare == value
            )
        ]
        if len(source) == 0:
            raise ValueNotInEnum(prop, value=value)
        elif len(source) == 1:
            return source[0]
        else:
            return source
    else:
        return value



from spinta import commands
from spinta.components import Context
from spinta.backends.fs.components import FileSystem


@commands.bootstrap.register(Context, FileSystem)
def bootstrap(context: Context, backend: FileSystem):
    pass



from geoalchemy2.types import Geometry
from frictionless.plugins.sql import SqlStorage


class GeoSqlStorage(SqlStorage):

    def _SqlStorage__read_convert_type(self, sql_type=None):
        if isinstance(sql_type, Geometry):
            return 'geometry'

        return super()._SqlStorage__read_convert_type(sql_type)



from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.backends.mongo.components import Mongo


@commands.migrate.register(Context, Manifest, Mongo)
def migrate(context: Context, manifest: Manifest, backend: Mongo):
    pass



from spinta import commands
from spinta.components import Context
from spinta.datasets.components import Dataset, Resource


@commands.check.register()
def check(context: Context, dataset: Dataset):
    for resource in dataset.resources.values():
        commands.check(context, resource)


@commands.check.register()
def check(context: Context, resource: Resource):
    pass



import functools
import hashlib
import re
from typing import Any
from typing import Dict

scope_re = re.compile(r'[^a-z0-9]+', flags=re.IGNORECASE)


@functools.lru_cache(maxsize=1000)
def _sanitize(scope: str):
    return scope_re.sub('_', scope).lower()


def name_to_scope(
    template: str,
    name: str,
    *,
    maxlen: int = None,
    params: Dict[str, Any] = None,
) -> str:
    """Return scope by given template possibly shortened on name part.
    """
    scope = template.format(name=name, **params)
    if maxlen and len(scope) > maxlen:
        surplus = len(scope) - maxlen
        name = name[:len(name) - surplus - 8] + hashlib.sha1(name.encode()).hexdigest()[:8]
        scope = template.format(name=name, **params)
    return _sanitize(scope)





from spinta import commands
from spinta.components import Context, MetaData
from spinta.manifests.components import Manifest


@commands.link.register(Context, Manifest)
def link(context: Context, manifest: Manifest):
    for nodes in manifest.objects.values():
        for node in nodes.values():
            commands.link(context, node)


@commands.link.register(Context, MetaData)
def link(context: Context, node: MetaData):
    pass



import operator
import itertools

from spinta.components import Context, Action, UrlParams, Model
from spinta.manifests.components import Manifest
from spinta.formats.components import Format
from spinta.formats.ascii.helpers import draw
from spinta.utils.nestedstruct import flatten
from spinta.formats.helpers import get_model_tabular_header


class Ascii(Format):
    content_type = 'text/plain'
    accept_types = {
        'text/plain',
    }
    params = {
        'width': {'type': 'integer'},
        'colwidth': {'type': 'integer'},
    }

    def __call__(
        self,
        context: Context,
        model: Model,
        action: Action,
        params: UrlParams,
        data,
        width=None,
        colwidth=42,
    ):
        manifest: Manifest = context.get('store').manifest

        data = iter(data)
        peek = next(data, None)

        if peek is None:
            return

        data = itertools.chain([peek], data)

        if '_type' in peek:
            groups = itertools.groupby(data, operator.itemgetter('_type'))
        else:
            groups = [(None, data)]

        for name, group in groups:
            if name:
                yield f'\n\nTable: {name}\n'
                model = manifest.models[name]

            rows = flatten(group)
            cols = get_model_tabular_header(context, model, action, params)

            if colwidth:
                width = len(cols) * colwidth

            buffer = [cols]
            tnum = 1

            for row in rows:
                buffer.append([row.get(c) for c in cols])
                if len(buffer) > 100:
                    yield from draw(buffer, name, tnum, width)
                    buffer = [cols]
                    tnum += 1

            if buffer:
                yield from draw(buffer, name, tnum, width)



from typing import Optional

import datetime
import uuid
import hashlib
import decimal
from uuid import UUID

import msgpack
import sqlalchemy as sa

from spinta import commands
from spinta.cli.helpers.data import ensure_data_dir
from spinta.components import Config
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.datasets.keymaps.components import KeyMap


class SqlAlchemyKeyMap(KeyMap):
    dsn: str = None

    def __init__(self, dsn: str = None):
        self.dsn = dsn
        self.engine = None
        self.metadata = None
        self.conn = None

    def __enter__(self):
        assert self.dsn is not None
        assert self.conn is None
        self.conn = self.engine.connect()
        return self

    def __exit__(self, *exc):
        self.conn.close()
        self.conn = None

    def get_table(self, name):
        if name not in self.metadata.tables:
            table = sa.Table(
                name, self.metadata,
                sa.Column('key', sa.Text, primary_key=True),
                sa.Column('hash', sa.Text, unique=True),
                sa.Column('value', sa.LargeBinary),
            )
            table.create(checkfirst=True)
        return self.metadata.tables[name]

    def encode(self, name: str, value: object) -> Optional[str]:
        # Make value msgpack serializable.
        if isinstance(value, (list, tuple)):
            value = [_encode_value(k) for k in value if k is not None]
            if len(value) == 0:
                return None
        else:
            if value is None:
                return None
            value = _encode_value(value)

        # Get value hash.
        value = msgpack.dumps(value, strict_types=True)
        hash = hashlib.sha1(value).hexdigest()

        # Try to find key by value hash.
        table = self.get_table(name)
        query = sa.select([table.c.key]).where(table.c.hash == hash)
        key = self.conn.execute(query).scalar()
        if key is None:

            # Create new key.
            key = str(uuid.uuid4())
            self.conn.execute(table.insert(), {
                'key': key,
                'hash': hash,
                'value': value,
            })

        return key

    def decode(self, name: str, key: str) -> object:
        table = self.get_table(name)
        query = sa.select([table.c.value]).where(table.c.key == key)
        value = self.conn.execute(query).scalar()
        value = msgpack.loads(value, raw=False)
        return value


def _encode_value(value):
    if isinstance(value, (datetime.datetime, datetime.date)):
        return value.isoformat()
    if isinstance(value, decimal.Decimal):
        return float(value)
    if isinstance(value, tuple):
        return list(value)
    if isinstance(value, UUID):
        return str(value)
    return value


@commands.configure.register(Context, SqlAlchemyKeyMap)
def configure(context: Context, keymap: SqlAlchemyKeyMap):
    rc: RawConfig = context.get('rc')
    config: Config = context.get('config')
    dsn = rc.get('keymaps', keymap.name, 'dsn', required=True)
    ensure_data_dir(config.data_path)
    dsn = dsn.format(data_dir=config.data_path)
    keymap.dsn = dsn


@commands.prepare.register(Context, SqlAlchemyKeyMap)
def prepare(context: Context, keymap: SqlAlchemyKeyMap):
    keymap.engine = sa.create_engine(keymap.dsn)
    keymap.metadata = sa.MetaData(keymap.engine)
